"""coBib's search result match object.

This simple data container combines the text from a search result with span(s) indicating the
matching position(s).
"""

from typing import NamedTuple

from rich.text import Text

from cobib.config import config


class Span(NamedTuple):
    """A span to indicate a substring positionally within a larger one."""

    start: int
    """The start of the spanned substring."""

    end: int
    """The end of the spanned substring."""


class Match(NamedTuple):
    """A match object combining a text and matching substring(s)."""

    text: str
    """The text of a search result."""

    spans: list[Span]
    """The spans where a matching substring was found."""

    def stylize(self) -> Text:
        """Return a stylized `rich.Text` of this match."""
        text = Text(self.text)
        for span in self.spans:
            text.stylize(
                config.theme.search.query,
                span.start,
                span.end,
            )
        return text



"""coBib's example configuration command.

This command can be used to extract the default (or example) configuration for the installed coBib
version.

You can run this command via:
```
cobib _example_config
```
"""

from __future__ import annotations

import argparse

from rich.console import ConsoleRenderable
from rich.syntax import Syntax
from typing_extensions import override

from cobib.commands.base_command import Command
from cobib.utils.rel_path import RelPath


class ExampleConfigCommand(Command):
    """The example configuration Command.

    This command does not take any arguments.
    """

    name = "example_config"

    @override
    def __init__(self, *args: str) -> None:
        super().__init__(*args)

        self._contents: list[str] = []

    @override
    @classmethod
    def init_argparser(cls) -> None:
        cls.argparser = argparse.ArgumentParser(prog="example_config", exit_on_error=True)

    @override
    def execute(self) -> None:
        root = RelPath(__file__).parent.parent
        with open(root / "config/example.py", "r", encoding="utf-8") as file:
            self._contents = [line.strip() for line in file.readlines()]

    @override
    def render_porcelain(self) -> list[str]:
        return self._contents

    @override
    def render_rich(self, *, background_color: str = "default") -> ConsoleRenderable:
        syntax = Syntax(
            "\n".join(self._contents), "python", background_color=background_color, word_wrap=False
        )
        return syntax



"""coBib's list results viewer widget.

This widget gets produces by `cobib.commands.list.ListCommand.render_textual`.
It subclasses textual's built-in `DataTable` widget.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from typing import ClassVar

from rich.text import Text
from textual.binding import Binding
from textual.coordinate import Coordinate
from textual.geometry import Region, clamp
from textual.widgets import DataTable
from textual.widgets.data_table import RowDoesNotExist
from typing_extensions import override

from cobib.config import config

from .motion_key import MotionKey


class ListView(DataTable[Text]):
    """coBib's list results viewer widget."""

    id = "cobib-list-view"

    BINDINGS: ClassVar[list[Binding | tuple[str, str] | tuple[str, str, str]]] = [
        Binding("j", "motion('down', 'cursor_down')", "Down", show=False),
        Binding("k", "motion('up', 'cursor_up')", "Up", show=False),
        Binding("h", "motion('left', 'cursor_left')", "Left", show=False),
        Binding("l", "motion('right', 'cursor_right')", "Right", show=False),
        Binding("down", "motion('down', 'cursor_down')", "Down", show=False),
        Binding("up", "motion('up', 'cursor_up')", "Up", show=False),
        Binding("left", "motion('left', 'cursor_left')", "Left", show=False),
        Binding("right", "motion('right', 'cursor_right')", "Right", show=False),
        Binding("home", "motion('home', 'scroll_home')", "Home", show=False),
        Binding("end", "motion('end', 'scroll_end')", "End", show=False),
        Binding("pageup", "motion('pageup', 'page_up')", "Page Up", show=False),
        Binding("pagedown", "motion('pagedown', 'page_down')", "Page Down", show=False),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | space | Toggle the expand/collapsed space of the current item. |
    | j, down | Moves one row down. |
    | k, up | Moves one row up. |
    | h, left | Moves to the left. |
    | l, right | Moves to the right. |
    | PageDown | Moves one page down. |
    | PageUp | Moves one page up. |
    | End | Moves to the bottom of the list. |
    | Home | Moves to the top of the list. |
    """

    DEFAULT_CSS = """
        ListView {
            height: 1fr;
        }
    """

    @property
    @override
    def scrollable_content_region(self) -> Region:
        # we shrink the scroll-able content region by the scroll offset at the bottom
        # NOTE: shrinking at the top does not seem to yield the intended result as it merely adds to
        # the offset at the bottom
        return super().scrollable_content_region.shrink((0, 0, config.tui.scroll_offset, 0))

    @override
    def validate_scroll_y(self, value: float) -> float:
        if not self.vertical_scrollbar.grabbed:
            # we implement the scroll offset at the top by forcing the scroll y value to be at
            # least an offset away from the cursor row
            if self.cursor_row - value < config.tui.scroll_offset:
                value = self.cursor_row - config.tui.scroll_offset
        return clamp(value, 0, self.row_count)

    @override
    def on_mount(self) -> None:
        self.fixed_columns = 1
        self.zebra_stripes = True
        self.cursor_type = "row"

    def action_motion(self, key: str, action: str) -> None:
        """Action to move the cursor.

        Under the hood, this delegates to the respective built-in cursor motion methods.
        However, this method also posts a `cobib.ui.components.motion_key.MotionKey` event.

        Args:
            key: the key which was pressed.
            action: the built-in action to trigger.
        """
        func = getattr(self, f"action_{action}", None)
        if func is not None:
            func()
        self.post_message(MotionKey(key))

    def get_current_label(self) -> str | None:
        """Gets the label of the entry currently under the cursor.

        Returns:
            The label of the entry currently under the cursor.
        """
        label = self.coordinate_to_cell_key(Coordinate(self.cursor_row, 0)).row_key.value
        return label

    def jump_to_label(self, label: str) -> None:
        """Jumps to the requested label.

        Args:
            label: the label to jump to.

        Raises:
            KeyError: when the label was not found in the current view.
        """
        try:
            idx = self.get_row_index(label)
        except RowDoesNotExist as exc:
            raise KeyError from exc
        self.move_cursor(row=idx, animate=True)



"""coBib's terminal user interface.

This class implements a beautiful TUI for coBib, leveraging
[`textual`](https://github.com/textualize/textual).

.. warning::

   This module makes no API stability guarantees! With it being based on
   [`textual`](https://textual.textualize.io/) which is still in very early stages of its
   development, breaking API changes in this module might be released as part of coBib's feature
   releases. You have been warned.
"""

from __future__ import annotations

import asyncio
import io
import logging
import shlex
from collections.abc import Awaitable, Callable, Coroutine
from contextlib import redirect_stderr, redirect_stdout
from inspect import iscoroutinefunction
from typing import TYPE_CHECKING, Any, ClassVar, cast

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.keys import Keys
from textual.logging import TextualHandler
from textual.screen import Screen
from textual.widgets import Footer, Header, Input, Static
from typing_extensions import override

from cobib.config import config
from cobib.database import Database
from cobib.ui.components import (
    EntryView,
    HelpScreen,
    InputScreen,
    ListView,
    LoggingHandler,
    LogScreen,
    MainContent,
    MotionKey,
    PresetFilterScreen,
    SearchView,
    SelectionFilter,
)
from cobib.utils.prompt import Confirm

from .ui import UI

if TYPE_CHECKING:
    from cobib import commands

LOGGER = logging.getLogger(__name__)


# NOTE: pylint and mypy are unable to understand that the `App` interface actually implements `run`


class TUI(UI, App[None]):  # type: ignore[misc]
    """The TUI class.

    This class does not support any extra command-line arguments compared to the base class.
    However, it also extends the `textual.app.App` interface. It can be used with the debugging
    tools of textual but starting it requires you to use the `-c` (a.k.a. `--command`) argument of
    `textual run` since `cobib` is installed as a command-line script.

    In short, here is how you can start the TUI using textual:
    ```
    textual run -c cobib
    ```
    This assumes that you are at the root of the cobib development folder. You can include
    additional command-line arguments after `cobib`, but when doing so you will need to put the
    entire command in quotes, like this:
    ```
    textual run -c "cobib -v"
    ```

    Adding the `--dev` argument to `textual run` will connect it to the debugging console which you
    can start in a separate shell via `textual console`.

    For more information on how to debug a textual app, please refer to
    [their documentation](https://textual.textualize.io/guide/devtools/).
    """

    DEFAULT_CSS = """
        Screen {
            layers: default;
            align-vertical: bottom;
        }
    """

    _PRESET_FILTER_BINDINGS: ClassVar[list[tuple[str, str, str]]] = [
        (f"{i}", f"preset_filter({i})", f"Preset #{i}") for i in range(10)
    ]
    BINDINGS: ClassVar[list[Binding | tuple[str, str] | tuple[str, str, str]]] = [
        ("q", "quit", "Quit"),
        ("question_mark", "push_screen('help')", "Help"),
        ("underscore", "toggle_layout", "Layout"),
        ("colon", "prompt(':')", "Prompt"),
        ("v", "select", "Select"),
        ("slash", "prompt('/')", "Search"),
        ("a", "prompt('add ')", "Add"),
        ("c", "prompt('review ', False, True)", "Review"),
        ("d", "delete", "Delete"),
        ("e", "edit", "Edit"),
        ("f", "filter", "Filter"),
        ("i", "prompt('import ')", "Import"),
        ("m", "prompt('modify ', False, True)", "Modify"),
        ("o", "open", "Open"),
        ("p", "preset_filter()", "Preset"),
        ("r", "prompt('redo', True)", "Redo"),
        ("s", "sort", "Sort"),
        ("u", "prompt('undo', True)", "Undo"),
        ("x", "prompt('export ', False, True)", "Export"),
        ("z", "push_screen('log')", "Log"),
        *_PRESET_FILTER_BINDINGS,
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | q | Quit's coBib. |
    | ? | Toggles the help screen. |
    | _ | Toggles between the horizontal and vertical layout. |
    | : | Starts the prompt for an arbitrary coBib command. |
    | v | Selects the current entry. |
    | / | Searches the database for the provided string. |
    | digit | Immediately selects the preset filter given by that digit (0 = reset). |
    | a | Prompts for a new entry to be added to the database. |
    | d | Deletes the current (or selected) entries. |
    | e | Edits the current entry. |
    | f | Allows filtering the table using `++/--` keywords. |
    | i | Imports entries from another source. |
    | m | Prompts for a modification (respects selection). |
    | o | Opens the current (or selected) entries. |
    | p | Allows selecting a preset filter (see `config.tui.preset_filters`). |
    | r | Redoes the last undone change. Requires git-tracking! |
    | s | Prompts for the field to sort by (use -r to list in reverse). |
    | u | Undes the last change. Requires git-tracking! |
    | x | Exports the current (or selected) entries. |
    | z | Toggles the log screen. |
    """

    SCREENS: ClassVar[dict[str, Screen[Any] | Callable[[], Screen[Any]]]] = {
        "help": HelpScreen,
        "log": LogScreen,
        "input": InputScreen,
        "preset_filter": PresetFilterScreen,
    }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializes the TUI.

        Args:
            *args: any positional arguments for textual's underlying `App` class.
            **kwargs: any keyword arguments for textual's underlying `App` class.
        """
        super().__init__(*args, **kwargs)
        self.console.push_theme(config.theme.build())
        self.root_logger.addHandler(TextualHandler())
        self.title = "coBib"
        self.sub_title = "The Console Bibliography Manager"
        self._list_args = ["-r"]
        self._filter: SelectionFilter = SelectionFilter()
        self._filters.append(self._filter)
        self._background_tasks: set[asyncio.Task] = set()  # type: ignore[type-arg]
        LoggingHandler(self, level=logging.INFO)

    @override
    def compose(self) -> ComposeResult:
        with MainContent(initial=ListView.id):
            yield ListView()
            yield SearchView(".")

        yield EntryView()

        yield Header()
        yield Footer()

    # Event reaction methods

    def on_motion_key(self, event: MotionKey) -> None:
        """Triggers on the custom `cobib.ui.components.motion_key.MotionKey` event.

        This function will update the entry shown in the `cobib.ui.components.entry_view.EntryView`
        widget. It only triggers on vertical motion keys.

        Args:
            event: the motion event.
        """
        if event.key in {Keys.Down, Keys.Up, Keys.PageDown, Keys.PageUp, Keys.Home, Keys.End}:
            self._show_entry()

    async def on_mount(self) -> None:
        """Triggers on the [`Mount`][1] event.

        This method takes care of initializing the desired layout and seeds the `EntryView` widget.

        [1]: https://textual.textualize.io/api/events/#textual.events.Mount
        """
        self.screen.styles.layout = "horizontal"
        await self._update_table()
        self._show_entry()

    # Action methods

    async def action_quit(self) -> None:
        """Action to display the quit dialog."""

        async def _prompt_quit() -> None:
            res = await Confirm.ask("Are you sure you want to quit?", default=True)
            if res:
                self.exit()

        task = asyncio.create_task(_prompt_quit())
        self._background_tasks.add(task)
        task.add_done_callback(self._background_tasks.discard)

    def action_toggle_layout(self) -> None:
        """The layout toggling action.

        This action will toggle between a horizontal and vertical layout.
        """
        layout = self.screen.styles.layout
        if layout is None:
            return
        main = self.query_one(MainContent)
        if layout.name.lower().startswith("horizontal"):
            self.screen.styles.layout = "vertical"
            main.styles.height = "2fr"
            main.styles.width = "1fr"
        elif layout.name.lower().startswith("vertical"):
            self.screen.styles.layout = "horizontal"
            main.styles.width = "2fr"
            main.styles.height = "1fr"
        self.refresh(layout=True)

    async def action_prompt(
        self, value: str, submit: bool = False, check_selection: bool = False
    ) -> None:
        """The prompt action.

        This action triggers an interactive user input via which an arbitrary coBib command can be
        executed as if it were run via the CLI.

        All commands which do not have an explicit action associated with them will go through this
        prompt action.

        Args:
            value: the string with which to pre-populate the input field.
            submit: if set, the user will not be prompted and the input will be submitted
                immediately (i.e. `value` gets parsed directly).
            check_selection: whether or not to check for a visual selection.
        """
        if check_selection and self._filter.selection:
            self._filter.active = True
            value += "-s "

        if submit:
            await self._process_input(value)
        else:
            await_mount = self.push_screen("input", self._process_input)
            inp_screen = cast(InputScreen, self.get_screen("input"))
            await await_mount
            inp = inp_screen.query_one(Input)
            inp.value = value
            inp.action_end()

    async def action_select(self) -> None:
        """The selection action.

        This action adds the entry currently under the cursor to the visual selection.
        """
        main = self.query_one(MainContent)
        label = main.get_current_label()

        if label in self._filter.selection:
            self._filter.selection.remove(label)
        else:
            self._filter.selection.add(label)

        main.notify_style_update()
        main.refresh()
        self.query_one(EntryView).query_one(Static).refresh()

    def action_delete(self) -> None:
        """The delete action.

        This action triggers the `cobib.commands.delete.DeleteCommand`.
        """
        main = self.query_one(MainContent)
        labels: list[str]
        if self._filter.selection:
            labels = list(self._filter.selection)
            self._filter.selection.clear()
        else:
            labels = [main.get_current_label()]

        self._run_command(["delete", *labels])

    async def action_edit(self) -> None:
        """The edit action.

        This action triggers the `cobib.commands.edit.EditCommand`.
        """
        self._edit_entry()
        await self._update_table()

    def action_open(self) -> None:
        """The open action.

        This action triggers the `cobib.commands.open.OpenCommand`.
        """
        main = self.query_one(MainContent)
        labels: list[str]
        if self._filter.selection:
            labels = list(self._filter.selection)
            self._filter.selection.clear()
        else:
            labels = [main.get_current_label()]

        self._run_command(["open", *labels])

    async def action_preset_filter(self, idx: int | None = None) -> None:
        """The preset filter selection action.

        This action selects (or prompts for) a preset filter from `config.tui.preset_filters`.

        Args:
            idx: the index of which preset filter to select. When this is `None`, the user will be
                prompted interactively. When `0`, all filters will be reset.
        """
        if idx is None:
            await_mount = self.push_screen("preset_filter", self._process_input)
            await await_mount
        else:
            idx = int(idx)
            if idx == 0:
                await self._process_input("list -r")
            elif idx <= len(config.tui.preset_filters):
                await self._process_input(f"list {config.tui.preset_filters[idx-1]}")

    async def action_filter(self) -> None:
        """The filter action.

        This action triggers the `cobib.commands.list.ListCommand` via the `action_prompt` and
        allows the user to provide additional filters.
        """
        await self.action_prompt("list " + " ".join(self._list_args) + " ")

    async def action_sort(self) -> None:
        """The sort action.

        This action triggers the `cobib.commands.list.ListCommand` via the `action_prompt` and
        allows the user to provide a field to be sorted by. This command is handled specially
        because it ensures that only a single sort argument can be given at a time.
        """
        try:
            # first, remove any previously used sort argument
            sort_arg_idx = self._list_args.index("-s")
            # NOTE: we need to pop twice in order to remove both: the `-s` option and the key which
            # was sorted by
            self._list_args.pop(sort_arg_idx)
            self._list_args.pop(sort_arg_idx)
        except ValueError:
            pass

        # add the sort option to the arguments
        self._list_args += ["-s"]

        await self.action_prompt("list " + " ".join(self._list_args) + " ")

    # Utility methods

    def _edit_entry(self) -> None:
        """Edits the current entry.

        Suspension of the App is handled directly by the edit command using
        `cobib.utils.context.get_active_app`.
        """
        main = self.query_one(MainContent)
        label = main.get_current_label()
        from cobib import commands

        commands.EditCommand(label).execute()

    def _show_entry(self, command: commands.ShowCommand | None = None) -> None:
        """Renders the entry currently under the cursor in the `EntryView` widget.

        Args:
            command: the `cobib.commands.show.ShowCommand` to execute. If this is `None`, a new
                instance will be constructed to show the entry at the current cursor location.
        """
        if command is None:
            main = self.query_one(MainContent)
            label = main.get_current_label()
            from cobib import commands

            command = commands.ShowCommand(label)
        command.execute()
        entry = self.query_one(EntryView)
        static = entry.query_one(Static)
        static.update(
            command.render_rich(
                background_color=entry.background_colors[1].rich_color.name,
            )
        )

    def _jump_to_entry(self, command: list[str]) -> None:
        """Jumps the cursor in the current view to the entry provided by the command arguments.

        Args:
            command: the list of command arguments to be passed to the
                `cobib.commands.show.ShowCommand`.
        """
        from cobib import commands

        show_cmd = commands.ShowCommand(*command)
        label = show_cmd.largs.label
        main = self.query_one(MainContent)
        try:
            main.jump_to_label(label)
        except KeyError:
            if label in Database():
                msg = (
                    f"The entry with label '{label}' exists in the database but not in the current "
                    "view. Displaying it only in the side panel."
                )
                LOGGER.info(msg)
        self._show_entry(show_cmd)

    async def _update_table(self) -> None:
        """Updates the list of entries displayed in the `MainContent`."""
        main = self.screen_stack[0].query_one(MainContent)
        old_table = main.query_one(ListView)
        from cobib import commands

        command = commands.ListCommand(*self._list_args)
        command.execute()
        table = command.render_textual()
        await main.replace_widget(table)
        table.focus()
        if old_table is not None:
            table.cursor_coordinate = old_table.cursor_coordinate
            table.scroll_x = old_table.scroll_x
            table.scroll_y = old_table.scroll_y
            del old_table
        self.refresh(layout=True)

    async def _update_tree(self, command: list[str]) -> None:
        """Updates the tree of search results displayed in the `MainContent`.

        Args:
            command: the list of command arguments to be passed to the
                `cobib.commands.search.SearchCommand`.
        """
        from cobib import commands

        subcmd = commands.SearchCommand(*command)
        await subcmd.execute()
        tree = subcmd.render_textual()
        main = self.query_one(MainContent)
        await main.replace_widget(tree)
        tree.focus()
        self.refresh(layout=True)

    async def _process_input(self, value: str) -> None:
        """Processes the input returned from the `cobib.ui.components.input_screen.InputScreen`.

        Args:
            value: the value put into the `Input` widget by the user.
        """
        if not value:
            return

        if value[0] == "/":
            value = "search " + value[1:]
        elif value[0] == ":":
            value = value[1:]

        command = shlex.split(value)

        if self._filter.active:
            command += ["--"]
            command.extend(list(self._filter.selection))
            self._filter.active = False

        if command and command[0]:
            if command[0].lower() == "init":
                LOGGER.error(
                    "You cannot run the 'init' command from within the TUI!\n"
                    "Please run this command outside the TUI."
                )
                return
            if command[0].lower() == "git":
                LOGGER.error(
                    "You cannot run the 'git' command from within the TUI because it is impossible "
                    "to foresee what the command might output or need for input etc.\n"
                    "Please run this command outside the TUI."
                )
                return
            if command[0].lower() == "show":
                self._jump_to_entry(command[1:])
            if command[0].lower() == "list":
                self._list_args = command[1:]
                await self._update_table()
            elif command[0].lower() == "search":
                await self._update_tree(command[1:])
            elif command[0].lower() == "edit":
                with self.suspend():
                    from cobib import commands

                    commands.EditCommand(*command[1:]).execute()
            else:
                self._run_command(command)

    @staticmethod
    async def _async_done_callback(
        task: Awaitable[None], async_callback: Callable[[], Coroutine[Any, Any, None]]
    ) -> None:
        """Adds an asynchronous callback for when the provided task is done.

        Args:
            task: the Task to await before running the next coroutine.
            async_callback: the asynchronous callback to run after the `task` has completed.
        """
        await task
        await async_callback()

    def _run_command(self, command: list[str]) -> None:
        """Parses and executes a cobib command with its arguments.

        This method also redirects `stdout` and `stderr` and captures their contents to be displayed
        as popups in the TUI.

        Args:
            command: a list of strings consisting of the command keyword and its arguments.
        """
        with redirect_stdout(io.StringIO()) as stdout:
            with redirect_stderr(io.StringIO()) as stderr:
                try:
                    from cobib import commands

                    subcmd = getattr(commands, command[0].title() + "Command")(*command[1:])

                    if not iscoroutinefunction(subcmd.execute):
                        subcmd.execute()
                    else:
                        # FIXME: in these cases, stdout and stderr cannot be captured

                        # 1. create subcommand execution task
                        task1 = asyncio.create_task(subcmd.execute())

                        # 2. create another task which chains an asynchronous callback after the
                        #    previous one
                        task2 = asyncio.create_task(
                            TUI._async_done_callback(task1, self._update_table)
                        )

                        # 3. ensure proper clean-up of all created tasks
                        self._background_tasks.add(task1)
                        task1.add_done_callback(self._background_tasks.discard)
                        self._background_tasks.add(task2)
                        task2.add_done_callback(self._background_tasks.discard)

                except SystemExit:
                    pass

        stdout_val = stdout.getvalue().strip()
        if stdout_val:
            self.notify(stdout_val, title="Output", severity="information", timeout=5)

        stderr_val = stderr.getvalue().strip()
        if stderr_val:
            self.notify(stderr_val, title="Error", severity="error", timeout=10)



"""coBib's Rich-Textual progress integration.

The `Progress` utility provided by this module either constructs a `rich.progress.Progress` object
or a custom textual widget to display a rich-progress indicator and update it accordingly. This was
favored over using textual's `textual.widgets.ProgressBar` to re-implement rich's more flexible
`rich.progress.Progress` displaying capabilities.
"""

from __future__ import annotations

from threading import RLock
from time import monotonic
from typing import Any

from rich.progress import Progress as RichProgress
from rich.progress import ProgressColumn, Task, TaskID
from textual.widgets import Static
from typing_extensions import override

from .context import get_active_app


class Progress:
    """A utility class to construct either a `rich` or `textual` progress indicator."""

    @staticmethod
    def initialize(*columns: str | ProgressColumn, **kwargs: Any) -> RichProgress:
        """Initializes a new progress indicator.

        When `get_active_app` returns a `textual` App, this will construct a `TextualProgress`
        widget, otherwise it falls back to constructing a `rich.progress.Progress` object.

        Args:
            columns: the columns to include in the progress indicator.
            kwargs: any keyword arguments.

        Returns:
            The new progress indicator.
        """
        app = get_active_app()
        if app is None:
            return RichProgress(*columns, **kwargs)
        return TextualProgress(*columns, **kwargs)


class TextualProgress(  # type: ignore[misc]
    Static, RichProgress, can_focus=False, can_focus_children=False
):
    """coBib's Frankenstein Rich-Textual progress widget."""

    DEFAULT_CSS = """
        TextualProgress {
            layout: horizontal;
            width: 100%;
            height: 1;
            offset-y: -1;
            dock: bottom;
        }
    """

    def __init__(
        self,
        *columns: str | ProgressColumn,
    ) -> None:
        """Initializes the Frankenstein Rich-Textual progress widget.

        Args:
            *columns: the columns to be reported. If this is empty, it will fall back to
                `get_default_columns`.
        """
        super().__init__()

        self.columns = columns or self.get_default_columns()

        self._lock = RLock()
        self._tasks: dict[TaskID, Task] = {}
        self._task_index: TaskID = TaskID(0)

        self.speed_estimate_period = 30.0
        self.get_time = monotonic

    @override
    def advance(self, task_id: TaskID, advance: float = 1) -> None:
        super().advance(task_id, advance)
        self.renderable = self.make_tasks_table(self.tasks)
        self.refresh()

    @override
    async def start(self) -> None:  # type: ignore[override]
        app = get_active_app()
        await_mount = app.mount(self)  # type: ignore[union-attr]
        await await_mount

    @override
    def stop(self) -> None:
        self.set_timer(5.0, self.remove)



"""coBib's search results viewer widget.

This widget gets produces by `cobib.commands.search.SearchCommand.render_textual`.
It subclasses textual's built-in `Tree` widget.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from typing import ClassVar, Union

from rich.text import Text
from textual.binding import Binding
from textual.widgets import Tree
from typing_extensions import override

from .motion_key import MotionKey


class SearchView(Tree[Union[str, Text]]):
    """coBib's search results viewer widget."""

    id = "cobib-search-view"

    BINDINGS: ClassVar[list[Binding | tuple[str, str] | tuple[str, str, str]]] = [
        Binding("enter", "toggle_all", "Toggle All"),
        Binding("j", "motion('down', 'cursor_down')", "Down", show=False),
        Binding("k", "motion('up', 'cursor_up')", "Up", show=False),
        Binding("h", "motion('left', 'cursor_left')", "Left", show=False),
        Binding("l", "motion('right', 'cursor_right')", "Right", show=False),
        Binding("down", "motion('down', 'cursor_down')", "Down", show=False),
        Binding("up", "motion('up', 'cursor_up')", "Up", show=False),
        Binding("left", "motion('left', 'cursor_left')", "Left", show=False),
        Binding("right", "motion('right', 'cursor_right')", "Right", show=False),
        Binding("home", "motion('home', 'scroll_home')", "Home", show=False),
        Binding("end", "motion('end', 'scroll_end')", "End", show=False),
        Binding("pageup", "motion('pageup', 'page_up')", "Page Up", show=False),
        Binding("pagedown", "motion('pagedown', 'page_down')", "Page Down", show=False),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | space | Toggle the expand/collapsed space of the current item. |
    | enter | Toggle the expand/collapsed space of the current item and all its children. |
    | j, down | Moves one row down. |
    | k, up | Moves one row up. |
    | h, left | Moves to the left. |
    | l, right | Moves to the right. |
    | PageDown | Moves one page down. |
    | PageUp | Moves one page up. |
    | End | Moves to the bottom of the tree. |
    | Home | Moves to the top of the tree. |
    """

    @override
    def on_mount(self) -> None:
        self.show_root = False

    def action_motion(self, key: str, action: str) -> None:
        """Action to move the cursor.

        Under the hood, this delegates to the respective built-in cursor motion methods.
        However, this method also posts a `cobib.ui.components.motion_key.MotionKey` event.

        Args:
            key: the key which was pressed.
            action: the built-in action to trigger.
        """
        func = getattr(self, f"action_{action}", None)
        if func is not None:
            func()
        self.post_message(MotionKey(key))

    def action_toggle_all(self) -> None:
        """Toggles the expansion of the current node and all of its children recursively."""
        if self.cursor_node is not None:
            self.cursor_node.toggle_all()

    def get_current_label(self) -> str | None:
        """Gets the label of the entry currently under the cursor.

        Returns:
            The label of the entry currently under the cursor.
        """
        if self.cursor_node is None:
            return None
        return str(self.cursor_node.data)

    def jump_to_label(self, label: str) -> None:
        """Jumps to the requested label.

        Args:
            label: the label to jump to.

        Raises:
            KeyError: when the label was not found in the current view.
        """
        for child in self.root.children:
            if child.data == label:
                self.select_node(child)
                self.scroll_to_node(child, animate=True)
                break
        else:
            raise KeyError



"""coBib's Review command.

This command allows you to review the entries of your database to quickly make changes as you go.
This is especially helpful as your knowledge of your database changes and develops with time.

In its simplest form, you can review your entire database in an interactive manner with the
following command:
```
cobib review
```

Since this can be quite a lot all at once, you can also narrow your review down using filters (see
also `cobib.commands.list_`) just like with some of the other commands. For example, the following
command reviews all entries from the year 2023:
```
cobib review -- ++year 2023
```
You can also specify a manual selection of labels (or visually select them in the TUI (see below)):
```
cobib review --selection -- Label1 Label2 ...
```

You can also narrow your review down to one or more fields of all the entries:
```
cobib review tags -- ++year 2023
cobib review ENTRYTYPE year
```
When restricting your review like the above, you will only get a preview of the selected set of
fields. If you require more context, you can either request that interactively (see below) or
provide the `--context` option like so:
```
cobib review --context ENTRYTYPE
```

### How it works

For every label that you review, you will be presented with its contents and asked to select an
action that you would like to take. Your choices are the following:

- `edit`: opens the entry in your `cobib.config.config.EditCommandConfig.editor` and allows you to
          freely modify it.
- `done`: indicates that you are happy with this entry and completes its review.
- `skip`: skips the review of the current entry indicating that you do not want to make any changes
          but also do not mark it as `done` (which means that when resuming a review, it will come
          up again).
- `context`: requests additional context and updates the preview to contain all the information
             stored in the entry. This action is only available when context has not been requested
             yet and when fields are being filtered in the first place.
- `finish`: closes the review process (even if some entries have not been reviewed yet) and saves
            your progress to the database. This allows you to perform reviews in smaller chunks
            because you can resume from the current review state at another time.
- `inline`: allows you to edit a specific field in-line (i.e. without having to open the entry in an
            external editor). When reviewing a single field, you can simply type `inline` to edit
            that one. When reviewing more than one, you need to specify which field you want to edit
            in-line like so: `inline <field>`. This is a highly experimental feature so please be
            aware of bugs and report any issues or suggestions online:
            https://gitlab.com/cobib/cobib/-/issues/new

### Resuming a previously started review

If you finished a review early (see `finish` above) you can easily continue where you left of with
the `--resume` option. This option takes a git commit identifier as its argument so you should
either find the SHA of the auto-commit of the review that you would like to continue (see also
`cobib.commands.git.GitCommand`), or if you know that it is (for example) the last thing that you
did, you can simply use `HEAD` as shown below:
```
cobib review --resume HEAD
```
In this way, you do not need to repeat any filters that you may have used and coBib will
automatically know which entries you have marked as `done` and will not show them to you again.
However, entries which you used `skip` on before, will now show up again.

### Additional options

There are some additional options to the `review` command.

You can indicate which entries of a review you have already completed with the `--done` argument:
```
cobib review --done Author2023 -- ++year 2023
```
The example above will review all entries from the year 2023, except for `Autho2023`. It is unlikely
that you will ever use this option manually if you have `cobib.config.config.DatabaseConfig.git`
enabled, because that will allow you to resume a previously started review more easily.


### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `c` key (as in "check") which will drop you into the prompt where you
can type out a normal command-line command:
```
:review <arguments go here>
```

.. note::
   If you have already selected one or more entries, the `--selection` argument will automatically
   be added.
"""

from __future__ import annotations

import argparse
import json
import logging
import subprocess
from collections.abc import Callable
from functools import wraps
from typing import cast

from rich.console import Group
from rich.prompt import InvalidResponse, PromptBase, PromptType
from rich.syntax import Syntax
from rich.text import Text
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database, Entry
from cobib.parsers import YAMLParser
from cobib.utils.git import is_inside_work_tree
from cobib.utils.prompt import Prompt
from cobib.utils.rel_path import RelPath

from .base_command import Command
from .edit import EditCommand
from .list_ import ListCommand

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class ReviewCommand(Command):
    """The Review Command.

    This command can parse the following arguments:

        * `field`: an optional list of fields to limit the review process to.
        * `-c`, `--context`: when specified, the full entries will be previewed even if selected
          fields have been provided. This disables the interactive `context` action by always
          enabling it.
        * `-r`, `--resume`: specifies a git commit SHA of a previous review command auto-commit from
          which to resume. This requires the git-integration to be enabled.
        * `-d`, `--done`: specifies a list of labels which will not be included in the review even
          if they would seamingly match the filter/selection. This is useful for continuing a
          previously unfinished review process.
        * `-s`, `--selection`: when specified, the positional arguments will *not* be
          interpreted as filters but rather as a direct list of entry labels. This can
          be used on the command-line but is mainly meant for the TUIs visual selection
          interface (hence the name).
        * in addition to the above, you can add `filters` to specify a subset of your
          database for reviewing. For more information refer to `cobib.commands.list_`.
    """

    name = "review"

    @override
    def __init__(self, *args: str) -> None:
        super().__init__(*args)

        if self.largs.done is None:
            # NOTE: we cannot make the default of the `done` argument an empty list because this
            # would remain the same list across multiple runs of the review command within the same
            # session.
            self.largs.done = []

        self.reviewed_entries: list[Entry] = []
        """A list of `cobib.database.Entry` objects which were marked as `done` by this review."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = argparse.ArgumentParser(
            prog="review", description="Review subcommand parser.", exit_on_error=True
        )
        parser.add_argument("field", type=str, nargs="*", help="the field(s) to review")
        parser.add_argument(
            "-c",
            "--context",
            action="store_true",
            help="When specified, the full entries will be previewed even if selected fields have "
            "been provided. This disables the interactive `context` action by always enabling it.",
        )
        parser.add_argument(
            "-r",
            "--resume",
            type=str,
            default=None,
            help="Specifies a git commit SHA of a previous review command auto-commit from which "
            "to resume. This requires the git-integration to be enabled.",
        )
        parser.add_argument(
            "-d",
            "--done",
            type=str,
            nargs="*",
            help="Specifies a list of labels which will not be included in the review even if they "
            "would seamlingly match the filter/selection. This is useful for continuing a "
            "previously unfinished review process.",
        )
        parser.add_argument(
            "-s",
            "--selection",
            action="store_true",
            help="When specified, the `filter` argument will be interpreted as a list of entry "
            "labels rather than arguments for the `list` command.",
        )
        parser.add_argument(
            "filter",
            nargs="*",
            help="You can specify filters as used by the `list` command in order to select a "
            "subset of labels to be reviewed. To ensure this works as expected you should add the "
            "pseudo-argument '--' before the list of filters. See also `list --help` for more "
            "information.",
        )
        cls.argparser = parser

    @override
    @classmethod
    def _parse_args(cls, args: tuple[str, ...]) -> argparse.Namespace:
        review_args = []
        filter_args = []
        found_sep = False
        for arg in args:
            if arg == "--":
                found_sep = True
                continue
            if found_sep:
                filter_args.append(arg)
            else:
                review_args.append(arg)

        largs = super()._parse_args(tuple(review_args))
        largs.filter = filter_args
        return largs

    @override
    async def execute(self) -> None:  # type: ignore[override]  # noqa: PLR0912,PLR0915
        if self.largs.resume is not None:
            LOGGER.info(f"Trying to resume review from {self.largs.resume}.")
            git_tracked = config.database.git
            if not git_tracked:
                msg = (
                    "You must enable coBib's git-tracking in order to use the `--resume` option of "
                    "the `review` command.\nPlease refer to the documentation for more information "
                    "on how to set this up."
                )
                LOGGER.error(msg)
                return

            file = RelPath(config.database.file).path
            root = file.parent
            if not is_inside_work_tree(root):
                msg = (
                    "You have configured, but not initialized coBib's git-tracking."
                    "\nPlease consult `cobib init --help` for more information on how to do so."
                )
                LOGGER.error(msg)
                return

            try:
                sha_message = subprocess.check_output(
                    ["git", "-C", root, "log", "--format=%B", "-n1", self.largs.resume]
                )
            except subprocess.CalledProcessError:
                LOGGER.error(f"Could not find the requested git commit: '{self.largs.resume}'")
                return

            try:
                decoded_message = sha_message.decode()
                trimmed_message = "".join(decoded_message.splitlines()[1:])
            except KeyError:
                LOGGER.error(f"Could not trim the commit message: '{decoded_message}'")
                return

            try:
                message_args = json.loads(trimmed_message)
            except json.JSONDecodeError:
                LOGGER.error(
                    f"Could not extract arguments from the trimmed message: '{trimmed_message}'"
                )
                return

            LOGGER.info("Found the git-commit from which to resume.")
            for key, val in message_args.items():
                if key == "resume":
                    # we do not overwrite the resume argument
                    continue
                setattr(self.largs, key, val)

        LOGGER.debug("Starting Review command.")

        Event.PreReviewCommand.fire(self)

        if self.largs.selection:
            LOGGER.info("Selection given. Interpreting `filter` as a list of labels")
            labels = [label for label in self.largs.filter if label not in self.largs.done]
        else:
            LOGGER.debug("Gathering filtered list of entries to be reviewed.")
            listed_entries, _ = ListCommand(*self.largs.filter).execute_dull()
            labels = [entry.label for entry in listed_entries if entry.label not in self.largs.done]

        yml = YAMLParser()

        bib = Database()

        for label in labels:
            LOGGER.debug(f"Starting review of entry '{label}'")

            _continue = False
            context = self.largs.context and self.largs.field

            while not _continue:
                entry = bib[label]

                if context:
                    LOGGER.debug("Context has been requested so the fields are not filtered.")
                elif self.largs.field:
                    LOGGER.debug(
                        f"Limiting the review to the requested fields: '{self.largs.field}'"
                    )
                    entry = Entry(
                        label,
                        {key: val for key, val in entry.data.items() if key in self.largs.field},
                    )

                dump = yml.dump(entry)
                if dump is None:
                    LOGGER.error(f"The entry '{label}' could not be dumped as YAML. Skipping it.")
                    continue

                syntax = Syntax(dump, "yaml", background_color="default", word_wrap=True)

                prompt_text = "What action what you like to perform?"
                choices = ["done", "skip", "edit", "inline", "finish", "help"]
                if not context and self.largs.field:
                    choices.insert(0, "context")

                res = await Prompt.ask(
                    prompt_text,
                    choices=choices,
                    pre_prompt_message=syntax,
                    process_response_wrapper=self._wrap_prompt_process_response,
                )
                if res == "context":
                    LOGGER.info("Requesting more context.")
                    context = True
                elif res == "done":
                    LOGGER.info(f"Marking entry '{label}' as done.")
                    self.largs.done.append(label)
                    self.reviewed_entries.append(entry)
                    _continue = True
                elif res == "skip":
                    LOGGER.info(f"Skipping entry '{label}'.")
                    _continue = True
                elif res == "edit":
                    new_text = EditCommand.edit(dump)
                    if dump == new_text:
                        LOGGER.info("No changes detected.")
                    else:
                        parsed = yml.parse(new_text)
                        new_entry = next(iter(parsed.values()))

                        if self.largs.field:
                            new_entry.merge(bib[label], ours=True)

                        if new_entry.label != label:
                            LOGGER.error(
                                "Renaming entries as part of the review process is not supported!"
                            )
                        else:
                            bib.update({label: new_entry})
                elif res.startswith("inline"):
                    try:
                        _, inline = res.split()
                    except ValueError:
                        inline = self.largs.field[0]
                    LOGGER.info(f"Editing field '{inline}' in-line.")

                    prompt_text = f"Please provide the new value for the field '{inline}'"

                    warning = Text.from_markup(
                        "[yellow]WARNING: the inline editing is a highly experimental feature!\n"
                        "Be aware of bugs and proceed with care.\nPlease report any issues or "
                        "suggestions online: https://gitlab.com/cobib/cobib/-/issues/new"
                    )

                    res = await Prompt.ask(
                        prompt_text,
                        input_text=str(entry.data[inline]),
                        pre_prompt_message=Group(warning, syntax),
                    )
                    if res.isnumeric():
                        res = int(res)
                    entry.data[inline] = res
                    entry.merge(bib[label], ours=True)
                    bib.update({label: entry})
                    inline = None

                elif res == "finish":
                    LOGGER.info("Finishing review early.")
                    break

            else:
                # we did NOT break out of the while loop so we continue with the next label
                continue

            # we DID break out of the while loop indicating that we want to finish the review early
            break

        Event.PostReviewCommand.fire(self)

        bib.save()

        self.git(allow_empty=True)

    def _wrap_prompt_process_response(
        self,
        func: Callable[[PromptBase[PromptType], str], PromptType],
    ) -> Callable[[PromptBase[PromptType], str], PromptType]:
        """A method to wrap a `PromptBase.process_response` method.

        This method wraps a `PromptBase.process_response` method in order to handle a user's request
        for additional help.

        Args:
            func: the `PromptBase.process_response` method to be wrapped.

        Returns:
            The wrapped `PromptBase.process_response` method.
        """

        @override  # type: ignore[misc]
        @wraps(func)
        def process_response(prompt: PromptBase[PromptType], value: str) -> PromptType:
            return_value: PromptType
            try:
                return_value = func(prompt, value)
            except InvalidResponse as exc:
                if not value.startswith("inline"):
                    raise exc
                return_value = cast(PromptType, value)

            if return_value == "help":
                LOGGER.debug("User requested help.")
                help_text = (
                    "[yellow]You may perform any of the following actions:\n"
                    "    edit: open the current entry for editing\n"
                    "    skip: skip the current entry\n"
                    "    done: mark the current entry as done\n"
                    "  finish: finish the review early\n"
                    "  inline: EXPERIMENTAL edit a field in-line"
                )
                if prompt.choices is not None and "context" in prompt.choices:
                    help_text += "\n context: add additional context to the preview"
                raise InvalidResponse(help_text)

            elif value.startswith("inline"):
                num_args = len(value.split())
                num_fields = len(self.largs.field)
                if num_args > 2 or (num_args == 1 and num_fields > 1):  # noqa: PLR2004
                    help_text = (
                        "[red]You are reviewing more than one field at once.\n"
                        "For editing one of them in-line, please specify which one like so:\n"
                        "  inline <field>"
                    )
                    raise InvalidResponse(help_text)

            return return_value

        return process_response



r"""coBib's Search command.

This command allows you to search your database for one or more regex-interpreted queries.
While doing so, it uses the `cobib.config.config.SearchCommandConfig.grep` tool to search associated
files, too.

As a simple example, you can query for a simple author name like so:
```
cobib search Einstein
```
You can also control whether the search is performed case *in*sensitive.
This is done via the `cobib.config.config.SearchCommandConfig.ignore_case` setting which defaults to
being `False`.
Besides this setting, you can always overwrite its value on the command line with the
`--ignore-case` (`-i` for short) and `--no-ignore-case` (`-I` for short; since v4.1.0) options.
Providing these options takes precedence over your configuration value.
Thus, the following is *always* case *in*sensitive, irrespective of your configuration.
```
cobib search --ignore-case Einstein
```
And the following is *always* sensitive to case:
```
cobib search --no-ignore-case Einstein
```

By default, the search command will provide you with 1 line of context above and below the actual
matches. You can change this number of lines by setting the `--context` option:
```
cobib search --context 4 Einstein
```
You can also permanently change the default value via the
`cobib.config.config.SearchCommandConfig.context` setting.

Finally, you can also combine the search with coBib's filtering mechanism to narrow your search down
to a subset of your database:
```
cobib search Einstein -- ++year 2020
```
Note, that we use the auxiliary `--` argument to separate the filters from the actual arguments.
While this is not strictly necessary it helps to disambiguate the origin of the arguments.

.. note::

   You may also pass other arguments to the `list` command such as `--sort <field>`, `--reverse`, or
   `--limit`. However, these affect which list of entries is being searched over and not the search
   results directly. Thus, the `--limit` will provide an upper bound on the number of entries that
   is being searched and does _not_ indicate the maximum number of search results (since this can
   still be lower).


### Approximate searching

.. note::
   New in version v5.1.0

Besides case-insensitive searches (see above), coBib now also provides more means to perform
approximate searches.

The `--decode-latex` (`-l` for short) argument will convert LaTeX sequences to Unicode characters
(where possible). For example, a LaTeX-encoded Umlaut like `\"o` will become `ö`. This can
significantly simplify your search, for example:
```
cobib search --decode-latex Körper
```
will match entries that contain `K{\"o}rper`.
You can enable this behavior by default by setting `config.commands.search.decode_latex = True`.
If you have enabled this setting, you can temporarily overwrite it from the command-line with the
`--no-decode-latex` (`-L` for short) argument.

Additionally, you can convert Unicode characters to a close ASCII equivalent with `--decode-unicode`
(`-u` for short) which can simplify your search further. Reusing the example above, you can now
search for entries containing `K{\"o}rper` as follows:
```
cobib search --decode-latex --decode-unicode Korper
```
You can enable this behavior by default by setting `config.commands.search.decode_unicode = True`.
If you have enabled this setting, you can temporarily overwrite it from the command-line with the
`--no-decode-unicode` (`-U` for short) argument.

Finally, if you install the optional [`regex`](https://pypi.org/project/regex/) dependency you can
even perform fuzzy searches. For example, you can specify the amount of fuzzy errors to allow via
the `--fuzziness` (`-z` for short) argument. Reusing the sample example again, you can allow for
typos in your search like so:
```
cobib search --decode-latex --decode-unicode --fuzziness 2 Koprer
```
You can specify a default amount of fuzzy errors by setting `config.commands.search.fuzziness` to
the desired value.

### Associated files

The search will also be performed on any associated files of your entries.
You can configure the tool which is used to perform this search via the
`cobib.config.config.SearchCommandConfig.grep` setting (defaults to `grep`).

If you do not want to search through associated files, you can specify the `--skip-files` argument.

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `/` key.

.. note::
   For more information on the searching mechanisms see also `cobib.database.entry.Entry.search`.
"""

from __future__ import annotations

import argparse
import asyncio
import logging

from rich.console import ConsoleRenderable
from rich.text import Text
from rich.tree import Tree
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Entry
from cobib.ui.components import SearchView
from cobib.utils.match import Match
from cobib.utils.progress import Progress
from cobib.utils.regex import HAS_OPTIONAL_REGEX

from .base_command import Command
from .list_ import ListCommand

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class SearchCommand(Command):
    """The Search Command.

    This command can parse the following arguments:

        * `query`: the required positional argument corresponds to the regex-interpreted text which
          will be searched for. You may provide multiple separate queries which will be searched for
          independently.
        * `-i`, `--ignore-case`: if specified, the search will be case-**in**sensitive. This
          overwrites the `cobib.config.config.SearchCommandConfig.ignore_case` setting.
        * `-I`, `--no-ignore-case`: if specified, the search will be case-sensitive. This
          overwrites the `cobib.config.config.SearchCommandConfig.ignore_case` setting.
        * `-l`, `--decode-latex`: if specified, all LaTeX sequences will be decoded before searched.
          This overwrites the `cobib.config.config.SearchCommandConfig.decode_latex` setting.
        * `-L`, `--no-decode-latex`: if specified, LaTeX sequences will be left unchanged before
          searched. This overwrites the `cobib.config.config.SearchCommandConfig.decode_latex`
          setting.
        * `-u`, `--decode-unicode`: if specified, all Unicode characters will be decoded before
          searched. This overwrites the `cobib.config.config.SearchCommandConfig.decode_unicode`
          setting.
        * `-U`, `--no-decode-unicode`: if specified, Unicode characters will be left unchanged
          before searched. This overwrites the
          `cobib.config.config.SearchCommandConfig.decode_unicode` setting.
        * `-z`, `--fuzziness`: you can specify the number of fuzzy errors to allow for search
          matches. This requires the optional `regex` dependency to be installed.
        * `-c`, `--context`: you can specify the number of lines of "context" which is the number of
          lines before and after the actual match to be included in the output. This is similar to
          the `-C` option of `grep`. You can configure the default value via the
          `cobib.config.config.SearchCommandConfig.context` setting.
        * `--skip-files`: if specified, associated files will **not** be searched.
        * in addition to the above, you can add `filters` to narrow the search down to a subset of
          your database. For more information refer to `cobib.commands.list_`.
    """

    name = "search"

    @override
    def __init__(self, *args: str) -> None:
        super().__init__(*args)

        self.entries: list[Entry] = []
        """A filtered list of entries searched over by this command."""

        self.matches: list[list[Match]] = []
        """The search matches detected by this command. This is a nested list of the following
        structure: the first list level iterates over the entries; the second list level iterates
        over the matches of any given entry; the third list level iterates the (context) lines of
        any given match."""

        self.hits: int = 0
        """The number of search hits detected by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = argparse.ArgumentParser(
            prog="search", description="Search subcommand parser.", exit_on_error=True
        )
        parser.add_argument("query", type=str, nargs="+", help="text to search for")
        ignore_case_group = parser.add_mutually_exclusive_group()
        ignore_case_group.add_argument(
            "-i",
            "--ignore-case",
            action="store_true",
            default=None,
            help="ignore case for searching",
        )
        ignore_case_group.add_argument(
            "-I",
            "--no-ignore-case",
            dest="ignore_case",
            action="store_false",
            default=None,
            help="do NOT ignore case for searching",
        )
        latex_group = parser.add_mutually_exclusive_group()
        latex_group.add_argument(
            "-l",
            "--decode-latex",
            action="store_true",
            default=None,
            help="decode LaTeX sequences for searching",
        )
        latex_group.add_argument(
            "-L",
            "--no-decode-latex",
            dest="decode_latex",
            action="store_false",
            default=None,
            help="do NOT decode LaTeX sequences for searching",
        )
        unicode_group = parser.add_mutually_exclusive_group()
        unicode_group.add_argument(
            "-u",
            "--decode-unicode",
            action="store_true",
            default=None,
            help="decode Unicode characters for searching",
        )
        unicode_group.add_argument(
            "-U",
            "--no-decode-unicode",
            dest="decode_unicode",
            action="store_false",
            default=None,
            help="do NOT decode Unicode characters for searching",
        )
        parser.add_argument(
            "-z",
            "--fuzziness",
            type=int,
            default=config.commands.search.fuzziness,
            help=(
                "how many fuzzy errors to allow for search matches. This requires the optional "
                "`regex` dependency to be installed."
            ),
        )
        parser.add_argument(
            "-c",
            "--context",
            type=int,
            default=config.commands.search.context,
            help="number of context lines to provide for each match",
        )
        parser.add_argument(
            "--skip-files",
            action="store_true",
            default=None,
            help="do NOT search through associated files",
        )
        parser.add_argument(
            "filter",
            nargs="*",
            help="You can specify filters as used by the `list` command in order to select a "
            "subset of labels to be searched. To ensure this works as expected you should add the "
            "pseudo-argument '--' before the list of filters. See also `list --help` for more "
            "information.",
        )
        cls.argparser = parser

    @override
    @classmethod
    def _parse_args(cls, args: tuple[str, ...]) -> argparse.Namespace:
        search_args = []
        filter_args = []
        found_sep = False
        for arg in args:
            if arg == "--":
                found_sep = True
                continue
            if found_sep:
                filter_args.append(arg)
            else:
                search_args.append(arg)

        largs = super()._parse_args(tuple(search_args))
        largs.filter = filter_args

        if largs.fuzziness > 0 and not HAS_OPTIONAL_REGEX:  # pragma: no branch
            LOGGER.warning(  # pragma: no cover
                "Using the `--fuzziness` argument requires the optional `regex` dependency to be "
                "installed! Falling back to `fuzziness=0`."
            )
            largs.fuzziness = 0  # pragma: no cover

        return largs

    @override
    async def execute(self) -> None:  # type: ignore[override]
        LOGGER.debug("Starting Search command.")

        Event.PreSearchCommand.fire(self)

        self.entries, _ = ListCommand(*self.largs.filter).execute_dull()

        ignore_case = config.commands.search.ignore_case
        if self.largs.ignore_case is not None:
            ignore_case = self.largs.ignore_case
        LOGGER.debug("The search will be performed case %ssensitive", "in" if ignore_case else "")

        decode_latex = config.commands.search.decode_latex
        if self.largs.decode_latex is not None:
            decode_latex = self.largs.decode_latex
        LOGGER.debug("The search will%s decode all LaTeX sequences", "" if decode_latex else " NOT")

        decode_unicode = config.commands.search.decode_unicode
        if self.largs.decode_unicode is not None:
            decode_unicode = self.largs.decode_unicode
        LOGGER.debug(
            "The search will%s decode all Unicode characters", "" if decode_unicode else " NOT"
        )

        progress_bar = Progress.initialize()
        optional_awaitable = progress_bar.start()  # type: ignore[func-returns-value]
        if optional_awaitable is not None:
            await optional_awaitable

        task = progress_bar.add_task("Searching...", total=len(self.entries))

        for entry in self.entries.copy():
            progress_bar.advance(task, 1)
            await asyncio.sleep(0)

            matches = entry.search(
                self.largs.query,
                context=self.largs.context,
                skip_files=self.largs.skip_files,
                ignore_case=ignore_case,
                decode_unicode=decode_unicode,
                decode_latex=decode_latex,
                fuzziness=self.largs.fuzziness,
            )
            if not matches:
                self.entries.remove(entry)
                continue

            self.matches.append(matches)
            self.hits += len(matches)

            LOGGER.debug('Entry "%s" includes %d hits.', entry.label, len(matches))

        progress_bar.stop()

        Event.PostSearchCommand.fire(self)

    @override
    def render_porcelain(self) -> list[str]:
        output = []
        for entry, matches in zip(self.entries, self.matches):
            title = f"{entry.label}::{len(matches)}"
            output.append(title)

            for idx, match_ in enumerate(matches):
                prefix = f"{idx+1}::"
                for line in match_.text.splitlines():
                    output.append(prefix + line.strip())

        return output

    @override
    def render_rich(self) -> ConsoleRenderable:
        tree = Tree(".", hide_root=True)
        for entry, matches in zip(self.entries, self.matches):
            subtree = tree.add(
                Text.from_markup(
                    f"[search.label]{entry.markup_label()}[/search.label] - {len(matches)} match"
                    + ("es" if len(matches) > 1 else "")
                )
            )

            for idx, match_ in enumerate(matches):
                matchtree = subtree.add(str(idx + 1))
                for line in match_.stylize().split():
                    matchtree.add(line)

        return tree

    @override
    def render_textual(self) -> SearchView:
        tree = SearchView(".")
        for entry, matches in zip(self.entries, self.matches):
            data = entry.label
            subtree = tree.root.add(
                Text.from_markup(
                    f"[search.label]{entry.markup_label()}[/search.label] - {len(matches)} match"
                    + ("es" if len(matches) > 1 else "")
                ),
                data=data,
                expand=not config.tui.tree_folding[0],
            )

            for idx, match_ in enumerate(matches):
                matchtree = subtree.add(
                    str(idx + 1),
                    data=data,
                    expand=not config.tui.tree_folding[1],
                )
                for line in match_.stylize().split():
                    matchtree.add_leaf(line, data=data)

        return tree



"""coBib's Export command.

You can use this command to export your database.
As of now only two output formats are available:
* BibLaTex files
* Zip archives

The use case of the former is obvious and can be achieved by the following:
```
cobib export --bibtex my_database.bib
```
The latter case will collect all associated files of your database in a single Zip archive:
```
cobib export --zip my_references.zip
```
This is an important feature because coBib (by design) allows you to spread associated files across
your entire file system.
With this command you can gather them in a neat package for sharing or transferring.

You can also limit the export to a subset of your database in one of two ways:

1. through filters:
   ```
   cobib export --bibtex my_private_database.bib -- ++tags private
   ```

2. through a custom selection (using `--selection` or the short-hand option `-s`)
   ```
   cobib export --selection --bibtex some_other_database.bib -- Label1 Label2
   ```

.. note::
   While this latter case is usable via the command-line interface it is more a side-effect of the
   TUI integration which provides a visual selection (defaults to the `v` key).
   The proper and arguably more useful case is the first case using filters.

Since v3.2.0, coBib supports automatic Journal abbreviations. After configuring them as explained in
`cobib.config.config.UtilsConfig.journal_abbreviations` you can leverage them during exporting like
so:
```
cobib export --abbreviate --bibtex my_database.bib
# or
cobib export --abbreviate --dotless --bibtex my_database.bib
```

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `x` key which will drop you into the prompt where you can type out a
normal command-line command:
```
:export <arguments go here>
```

.. note::
   If you have already selected one or more entries, the `--selection` argument will automatically
   be added.
"""

from __future__ import annotations

import argparse
import logging
from zipfile import ZipFile

from typing_extensions import override

from cobib.config import Event
from cobib.database import Database, Entry
from cobib.parsers.bibtex import BibtexParser
from cobib.utils.journal_abbreviations import JournalAbbreviations
from cobib.utils.rel_path import RelPath

from .base_command import Command
from .list_ import ListCommand

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class ExportCommand(Command):
    """The Export Command.

    This command can parse the following arguments:

        * `-b`, `--bibtex`: specifies a BibLaTex filename into which to export.
        * `-z`, `--zip`: specifies a Zip-filename into which to export associated files.
        * `-a`, `--abbreviate`: abbreviate the Journal names before exporting. See also
          `cobib.config.config.UtilsConfig.journal_abbreviations`.
        * `--dotless`: remove punctuation from the Journal abbreviations.
        * `-s`, `--selection`: when specified, the positional arguments will *not* be
          interpreted as filters but rather as a direct list of entry labels. This can
          be used on the command-line but is mainly meant for the TUIs visual selection
          interface (hence the name).
        * in addition to the above, you can add `filters` to specify a subset of your
          database for exporting. For more information refer to `cobib.commands.list_`.
    """

    name = "export"

    @override
    def __init__(self, *args: str) -> None:
        super().__init__(*args)

        self.exported_entries: list[Entry] = []
        """A list of `cobib.database.Entry` objects which were exported by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = argparse.ArgumentParser(
            prog="export", description="Export subcommand parser.", exit_on_error=True
        )
        parser.add_argument(
            "-b", "--bibtex", type=argparse.FileType("a"), help="BibLaTeX output file"
        )
        parser.add_argument("-z", "--zip", type=argparse.FileType("a"), help="zip output file")
        parser.add_argument(
            "-s",
            "--selection",
            action="store_true",
            help="When specified, the `filter` argument will be interpreted as a list of entry "
            "labels rather than arguments for the `list` command.",
        )
        parser.add_argument(
            "filter",
            nargs="*",
            help="You can specify filters as used by the `list` command in order to select a "
            "subset of labels to be exported. To ensure this works as expected you should add the "
            "pseudo-argument '--' before the list of filters. See also `list --help` for more "
            "information.",
        )
        parser.add_argument(
            "-a", "--abbreviate", action="store_true", help="Abbreviate journal names"
        )
        parser.add_argument(
            "--dotless", action="store_true", help="Remove punctuation from journal abbreviations"
        )
        cls.argparser = parser

    @override
    def execute(self) -> None:
        LOGGER.debug("Starting Export command.")

        Event.PreExportCommand.fire(self)

        if self.largs.bibtex is None and self.largs.zip is None:
            msg = "No output file specified!"
            LOGGER.error(msg)
            return
        if self.largs.zip is not None:
            self.largs.zip = ZipFile(self.largs.zip.name, "w")

        if self.largs.selection:
            LOGGER.info("Selection given. Interpreting `filter` as a list of labels")
            labels = self.largs.filter
            bib = Database()
            for label in labels:
                try:
                    self.exported_entries.append(bib[label])
                except KeyError:
                    msg = f"No entry with the label '{label}' could be found."
                    LOGGER.warning(msg)
        else:
            LOGGER.debug("Gathering filtered list of entries to be exported.")
            self.exported_entries, _ = ListCommand(*self.largs.filter).execute_dull()

        bibtex_parser = BibtexParser()

        for entry in self.exported_entries:
            LOGGER.info('Exporting entry "%s".', entry.label)
            if self.largs.bibtex is not None:
                if self.largs.abbreviate and "journal" in entry.data.keys():
                    entry.data["journal"] = JournalAbbreviations.abbreviate(
                        entry.data["journal"], dotless=self.largs.dotless
                    )
                entry_str = bibtex_parser.dump(entry)
                self.largs.bibtex.write(entry_str)
            if self.largs.zip is not None:
                if "file" in entry.data.keys() and entry.file is not None:
                    files = entry.file
                    if not isinstance(files, list):
                        files = [files]  # type: ignore[unreachable]
                    for file in files:
                        path = RelPath(file).path
                        LOGGER.debug(
                            'Adding "%s" associated with "%s" to the zip file.', path, entry.label
                        )
                        self.largs.zip.write(path, path.name)

        Event.PostExportCommand.fire(self)

        if self.largs.bibtex is not None:
            self.largs.bibtex.close()
        if self.largs.zip is not None:
            self.largs.zip.close()



"""coBib's command to unify entry labels.

This is a convenience command to unify the labels of all your database entries to follow the value
set for `cobib.config.config.DatabaseFormatConfig.label_default`.
"""

from __future__ import annotations

import argparse
import contextlib
import logging
from io import StringIO

from rich.console import ConsoleRenderable
from rich.text import Text
from typing_extensions import override

from cobib.config import config

from .base_command import Command
from .modify import ModifyCommand

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class UnifyLabelsCommand(Command):
    """The label unification Command.

    This command can parse the following arguments:

        * `-a`, `--apply`: if specified, the label unification will actually be applied. The default
            is to run in "dry"-mode which only prints the modifications.
    """

    name = "unify_labels"

    @override
    def __init__(self, *args: str) -> None:
        super().__init__(*args)

        self._contents: list[str] = []

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = argparse.ArgumentParser(
            prog="unify_labels",
            description="Label unification subcommand parser.",
            exit_on_error=True,
        )
        parser.add_argument(
            "-a",
            "--apply",
            action="store_true",
            help="Actually apply the modifications rather than run in 'dry'-mode",
        )
        cls.argparser = parser

    @override
    def execute(self) -> None:
        modify_args = [
            "--dry",
            f"label:{config.database.format.label_default}",
            "--",
            # the following ensures that the command gets run on the entire database
            "--label",
            "some_non_existent_label_123456_abcdef",
        ]
        if self.largs.apply:
            modify_args = modify_args[1:]

        with contextlib.redirect_stderr(StringIO()) as out:
            cmd = ModifyCommand(*modify_args)
            cmd.execute()

        self._contents = out.getvalue().strip().split("\n")

    @override
    def render_porcelain(self) -> list[str]:
        return self._contents

    @override
    def render_rich(self) -> ConsoleRenderable:
        text = Text("\n".join(self._contents))
        text.highlight_words(["ERROR"], "bold red")
        text.highlight_words(["WARNING"], "bold yellow")
        text.highlight_words(["HINT"], "green")
        text.highlight_words(["INFO"], "blue")
        return text



"""coBib's Delete command.

This command can be used to deleted entries from the database.
```
cobib delete <label 1> [<label 2> ...]
```

When you delete an entry, the value of `cobib.config.config.DeleteCommandConfig.preserve_files`
setting (added in v4.1.0) determines whether the associated files will be deleted, too. This
defaults to `False`, meaning that they *will* be deleted. You can overwrite the value of this
setting at runtime with the `--preserve-files` and `--no-preserve-files` arguments, respectively.
I.e. the following will **not** delete your files:
```
cobib delete --preserve-files <label 1> [<label 2> ...]
```
While this command will always delete them:
```
cobib delete --no-preserve-files <label 1> [<label 2> ...]
```

As of coBib v4.1.0, the user will be asked to confirm the deletion via an interactive prompt. This
can be disabled by setting `cobib.config.config.DeleteCommandConfig.confirm` to `False`.

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `d` key.
"""

from __future__ import annotations

import argparse
import logging
import os

from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.prompt import Confirm
from cobib.utils.rel_path import RelPath

from .base_command import Command

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class DeleteCommand(Command):
    """The Delete Command.

    This command can parse the following arguments:

        * `labels`: one (or multiple) labels of the entries to be deleted.
        * `--preserve-files`: skips the deletion of any associated files. This overwrites the
          `cobib.config.config.DeleteCommandConfig.preserve_files` setting.
        * `--no-preserve-files`: does NOT skip the deletion of any associated files. This overwrites
          the `cobib.config.config.DeleteCommandConfig.preserve_files` setting.
    """

    name = "delete"

    @override
    def __init__(self, *args: str) -> None:
        super().__init__(*args)

        self.deleted_entries: set[str] = set()
        """A set of labels which were deleted by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = argparse.ArgumentParser(
            prog="delete", description="Delete subcommand parser.", exit_on_error=True
        )
        parser.add_argument("labels", type=str, nargs="+", help="labels of the entries")
        preserve_files_group = parser.add_mutually_exclusive_group()
        preserve_files_group.add_argument(
            "--preserve-files",
            action="store_true",
            default=None,
            help="do not delete associated files",
        )
        preserve_files_group.add_argument(
            "--no-preserve-files",
            dest="preserve_files",
            action="store_false",
            default=None,
            help="delete associated files",
        )
        cls.argparser = parser

    @override
    async def execute(self) -> None:  # type: ignore[override]
        LOGGER.debug("Starting Delete command.")

        Event.PreDeleteCommand.fire(self)

        preserve_files = config.commands.delete.preserve_files
        if self.largs.preserve_files is not None:
            preserve_files = self.largs.preserve_files
        LOGGER.info("Associated files will%s be preserved.", "" if preserve_files else " not")

        bib = Database()
        for label in self.largs.labels:
            try:
                LOGGER.debug("Attempting to delete entry '%s'.", label)

                if config.commands.delete.confirm:
                    prompt_text = f"Are you sure you want to delete the entry '{label}'?"

                    res = await Confirm.ask(prompt_text, default=True)
                    if not res:
                        continue

                entry = bib.pop(label)
                if not preserve_files:
                    for file in entry.file:
                        path = RelPath(file)
                        try:
                            LOGGER.debug("Attempting to remove associated file '%s'.", str(path))
                            os.remove(path.path)
                        except FileNotFoundError:
                            pass

                self.deleted_entries.add(label)
            except KeyError:
                pass

        Event.PostDeleteCommand.fire(self)
        bib.save()

        self.git()

        for label in self.deleted_entries:
            msg = f"'{label}' was removed from the database."
            LOGGER.info(msg)



"""coBib's database module.

coBib uses a plain-test YAML file to store the bibliographic database.
This module contains the components which represent this database at runtime.

.. warning::

   If you feel like coBib spends an awful lot of time reading in your database, you can try to
   improve performance by _linting your database_. Running `cobib _lint_database` will print any
   logging messages that occur during the database parsing, indicating the coBib had to apply some
   changes manually at runtime (which will happen every time you need to read-in the database). You
   can automatically fix all these linting suggestions based on how coBib automatically treats the
   entries internally by adding the `--format` argument: `cobib _lint_database -f`.

.. note::

   As of version 4.4.0 coBib also has a primitive caching mechanism which will significantly improve
   the startup performance. You can configure the location of the cache via the
   `cobib.config.config.DatabaseConfig.cache` setting or disable it entirely by changing this
   setting to `None`.
   Note, that linting your database will still improve startup performance when your cache becomes
   outdated for any reason (for example manual editing of the database file or syncing it via the
   git integration).
"""

from .author import Author as Author
from .database import Database as Database
from .entry import Entry as Entry



"""coBib's Database class."""

from __future__ import annotations

import logging
import pickle
import re
import sys
from collections import OrderedDict
from pathlib import Path
from typing import ClassVar, cast

from cobib.config import LabelSuffix, config
from cobib.utils.rel_path import RelPath

from .entry import Entry

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class Database(OrderedDict[str, Entry]):
    """coBib's Database class is a runtime interface to the plain-test YAML file.

    This class is a *singleton*.
    Thus, accessing `cobib.database.Database` will always yield the identical instance at runtime.
    This ensures data consistency during all operations on the database.
    """

    _instance: Database | None = None
    """The singleton instance of this class."""

    _unsaved_entries: ClassVar[dict[str, str | None]] = {}
    """A dictionary of changed entries which have not been written to the database file, yet.
    The keys are the entry labels. If it the entry was removed, the value of this key is `None`.
    Otherwise it is set to the label of the changed entry (which may be different from the previous
    label, indicating a renaming of the entry)."""

    _read: bool = False
    """Indicates whether the database has already been read. This state is purely used to avoid an
    endless recursion during the class construction. If this state if `False`, the `__new__` method
    will automatically call `read`. The `read` method then immediately sets this flag to `True`,
    preventing further recursion. This is especially important when dealing with the separate call
    to `read_cache`."""

    def __new__(cls, *, bypass_cache: bool = False) -> Database:
        """Singleton constructor.

        This method gets called when accessing `Database` and enforces the singleton pattern
        implemented by this class.

        Args:
            bypass_cache: whether or not to try reading the cache. Set this to `True` to bypass the
                cache no matter its age or the user configuration.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._unsaved_entries = {}
        if not cls._read:
            cls.read(bypass_cache=bypass_cache)
        return cls._instance

    def update(self, new_entries: dict[str, Entry]) -> None:  # type: ignore[explicit-override,override]
        """Updates the database with the given dictionary of entries.

        This function wraps `OrderedDict.update` and adds the labels of the changed entries to the
        dictionary of unsaved entries (`Database._unsaved_entries`). This will minimize IO access by
        only actually writing the unsaved entries in batches.

        Args:
            new_entries: the dictionary of labels mapping to entries which are to be written to the
                database.
        """
        for label in new_entries.keys():
            LOGGER.debug("Updating entry %s", label)
            Database._unsaved_entries[label] = label
        super().update(new_entries)

    def pop(self, label: str) -> Entry:  # type: ignore[explicit-override,override]
        """Pops the entry pointed to by the given label.

        This function wraps `OrderedDict.pop` and adds the removed labels to the unsaved entries
        (`Database._unsaved_entries`). This will minimize IO access by only actually writing the
        unsaved entries in batches.

        Args:
            label: the label of the entry to be removed.

        Returns:
            The entry pointed to by the given label.
        """
        entry: Entry = super().pop(label)
        LOGGER.debug("Removing entry: %s", label)
        Database._unsaved_entries[label] = None
        return entry

    def rename(self, old_label: str, new_label: str) -> None:
        """Renames an entry label.

        This function performs no actual changes to the database. It merely registers a rename
        operation in `Database._unsaved_entries`.

        Args:
            old_label: the previous label.
            new_label: the new label.
        """
        LOGGER.debug("Renaming entry '%s' to '%s'.", old_label, new_label)
        Database._unsaved_entries[old_label] = new_label
        if new_label != old_label:
            # NOTE: this is not technically needed but the rename method is exploited during
            # database linting with "fake" renames in order to register entries for re-writing
            # during saving
            super().pop(old_label)

    def disambiguate_label(self, label: str, entry: Entry) -> str:
        """Disambiguate a given label to ensure it becomes unique.

        This function ensures that a label is unique by appending a configurable suffix to a label
        if it is already present in the database at runtime.

        Args:
            label: the label which to disambiguate.
            entry: the `Entry` to which this label belongs.

        Returns:
            A unique label.
        """
        if label not in self.keys():
            LOGGER.info("The label '%s' does not yet exist in the runtime database.", label)
            return label

        if self[label] == entry:
            LOGGER.log(
                35,
                "Even though the label '%s' already exists in the runtime database, the entry is "
                "identical and, thus, no further disambiguation is necessary.",
                label,
            )
            return label

        LOGGER.warning("The label '%s' already exists in database. Running disambiguation.", label)
        separator, enumerator = config.database.format.label_suffix
        offset = 0
        while True:
            offset += 1
            new_label: str = label + separator + enumerator(offset)  # type: ignore[operator]
            if new_label not in self.keys():
                LOGGER.info("Found new unique label: %s", new_label)
                return new_label
            LOGGER.log(
                35,
                "The label '%s' also already exists in the database. You are seeing this because "
                "you are running a disambiguation of the label '%s'. You may want to check whether "
                "these two entries are related and (if so) edit or merge them manually. For more "
                "information see also: https://gitlab.com/cobib/cobib/-/issues/121",
                new_label,
                label,
            )

    def find_related_labels(self, label: str) -> tuple[set[str], set[str]]:
        """Finds related labels to the provided one.

        Args:
            label: the original label for which to find related ones.

        Returns:
            A pair of sets of strings. The first set contains directly related labels, i.e. ones
            which match the provided label exactly, modulo the value of their suffix based on
            `config.database.format.label_suffix`. The second set contains indirectly related
            labels, i.e. ones which also start with the same text as the original label but do not
            have a matching suffix.
        """
        separator, enumerator = config.database.format.label_suffix
        trimmed_label, _ = LabelSuffix.trim_label(label, separator, enumerator)

        directly_related_labels = set()
        indirectly_related_labels = set()
        for existing_label in self.keys():
            if not existing_label.startswith(trimmed_label):
                continue

            raw_label, _ = LabelSuffix.trim_label(existing_label, separator, enumerator)
            if raw_label == trimmed_label:
                directly_related_labels.add(existing_label)
            else:
                indirectly_related_labels.add(existing_label)

        return directly_related_labels, indirectly_related_labels

    @classmethod
    def reset(cls) -> None:
        """Resets the database.

        This clears the contents of the singleton instance and resets the `_read` class attribute.
        """
        if cls._instance is not None:
            cls._instance.clear()
        cls._read = False

    @classmethod
    def read(cls, *, bypass_cache: bool = False) -> None:
        """Reads the database file.

        The YAML database file pointed to by the configuration file is read in and parsed.
        This uses `cobib.parsers.YAMLParser` to parse the data.
        This function clears the contents of the singleton `Database` instance and resets
        `Database._unsaved_entries` to an empty dictionary. Thus, a call to this function
        *irreversibly* synchronizes the state of the runtime `Database` instance to the actually
        written contents of the database file on disc.

        Args:
            bypass_cache: whether or not to try reading the cache. Set this to `True` to bypass the
                cache no matter its age or the user configuration.
        """
        if cls._instance is None:
            cls(bypass_cache=bypass_cache)
            return
        _instance = cls._instance

        try:
            if bypass_cache:
                raise CacheError("Bypassing the cache.")
            Database.read_cache()
        except CacheError as exc:
            LOGGER.log(
                exc.log_level, "Encountered the following exception during cache lookup: '%s'", exc
            )

            file = RelPath(config.database.file).path
            try:
                LOGGER.info("Loading database file: %s", file)

                from cobib.parsers.yaml import YAMLParser

                cls._read = True
                _instance.clear()
                _instance.update(YAMLParser().parse(file))
            except FileNotFoundError:
                LOGGER.critical(
                    "The database file %s does not exist! Please run `cobib init`!", file
                )
                sys.exit(1)

            Database.save_cache()

        cls._unsaved_entries.clear()

    @classmethod
    def save(cls) -> None:
        """Saves all unsaved entries.

        This uses `cobib.parsers.YAMLParser` to save all entries in `Database._unsaved_entries` to
        disc. In doing so, this function preserves the order of the entries in the database file by
        overwriting changed entries in-place and appending new entries to the end of the file.

        The method of determining whether an entry was added, changed or removed is the following:
        1. we read in the current database as written to disc.
        2. we iterate all lines and determine the label of the entry we are currently on.
        3. if this label is not in `Database._unsaved_entries` we continue.
        4. Otherwise we query the runtime `Database` instance for the new contents of the unsaved
           (and therefore changed) entry and remove the label from `Database._unsaved_entries`.
        5. Using `Entry.save` and a `cobib.parsers.YAMLParser` we overwrite the previous lines of
           the changed entry.
        6. Finally, all labels still left in `Database._unsaved_entries` are newly added entries and
           can simply be appended to the file.

        In order to optimize performance and IO access, all of the above is done with a single call
        to `write`.
        """
        if cls._instance is None:
            cls()
        _instance = cast(Database, cls._instance)

        from cobib.parsers.yaml import YAMLParser

        yml = YAMLParser()

        file = RelPath(config.database.file).path
        with open(file, "r", encoding="utf-8") as bib:
            lines = bib.readlines()

        label_regex = re.compile(r"^([^:]+):$")

        overwrite = False
        cur_label: str = ""
        buffer: list[str] = []
        for line in lines:
            try:
                matches = label_regex.match(line)
                if matches is None:
                    raise AttributeError
                new_label = matches.groups()[0]
                if new_label in cls._unsaved_entries:
                    LOGGER.debug('Entry "%s" found. Starting to replace lines.', new_label)
                    overwrite = True
                    cur_label = new_label
                    continue
            except AttributeError:
                pass
            if overwrite and line.startswith("..."):
                LOGGER.debug('Reached end of entry "%s".', cur_label)
                overwrite = False

                new_label = cast(str, cls._unsaved_entries.pop(cur_label))
                entry = _instance.get(new_label, None)
                if entry:
                    LOGGER.debug('Writing modified entry "%s".', new_label)
                    entry_str = entry.save(parser=yml)
                    buffer.append("\n".join(entry_str.split("\n")[1:]))
                else:
                    # Entry has been deleted. Pop the previous `---` line.
                    LOGGER.debug('Deleting entry "%s".', new_label)
                    buffer.pop()
                # we pop `new_label` too, because in case of a rename it differs from `cur_label`
                if new_label is not None:
                    cls._unsaved_entries.pop(new_label, None)
            elif not overwrite:
                # keep previous line
                buffer.append(line)

        if cls._unsaved_entries:
            for label in cls._unsaved_entries.copy().values():
                if label is None:
                    # should never occur but we avoid a type exception
                    continue
                LOGGER.debug('Adding new entry "%s".', label)
                entry_str = _instance[label].save(parser=yml)
                buffer.append(entry_str)
                cls._unsaved_entries.pop(label)

        with open(file, "w", encoding="utf-8") as bib:
            for line in buffer:
                bib.write(line)

        Database.save_cache()

    @staticmethod
    def _get_cache_file() -> Path | None:
        """Returns the full path to the cache file for the current database file.

        Note, that this respects the `config.database.cache` and `config.database.file` settings.
        If the former is set to `None`, this method will return `None`, indicating that caching is
        disabled. The `config.database.file` setting is used to determine a unique name for the
        caching filename, simply by replacing all path separators by `%2f` (the ASCII hex code for
        the `/` symbol).


        """
        if config.database.cache is None:
            return None

        cache_location = RelPath(config.database.cache).path

        database_location = RelPath(config.database.file).path
        # we remove the anchor from the database path in order to get rid of any drive letter
        # (Windows) and root node which we do not want to include in the caching filename
        database_anchor = database_location.anchor
        database_relative = database_location.relative_to(database_anchor)
        # we can now join the parts of this relative path to form a new unique filename
        file_name = "%2f".join(database_relative.parts)

        cache_file = (cache_location / file_name).with_suffix(".pickle")
        return cache_file

    @staticmethod
    def _is_cache_outdated(cache_file: Path) -> bool:
        """Compares the age of the cache file with the database file itself.

        Args:
            cache_file: the full path to the cache file for the current database file.

        Returns:
            `True` if the cache file is outdated compared to the database file, `False` otherwise.

        Raises:
            FileNotFoundError: if the cache file does not exist.
        """
        if not cache_file.exists():
            raise FileNotFoundError

        database_location = RelPath(config.database.file).path
        cache_age = cache_file.stat().st_mtime
        database_age = database_location.stat().st_mtime

        return cache_age < database_age

    @classmethod
    def read_cache(cls) -> None:
        """Reads the database from a cache.

        Raises:
            CacheError: if caching is disable via `cobib.config.config.DatabaseConfig.cache`.
            CacheError: if the current database has not been cached yet.
            CacheError: if the cached database is older than the last modification time of the
                database file itself, indicating that the cache is outdated.
        """
        cache_file = cls._get_cache_file()
        if cache_file is None:
            exc = CacheError("caching is disabled via the configuration")
            exc.log_level = logging.DEBUG
            raise exc

        try:
            cache_outdated = cls._is_cache_outdated(cache_file)

            if cache_outdated:
                raise CacheError("the cached database is outdated")
        except FileNotFoundError as exc:
            raise CacheError("the database has not been cached yet") from exc

        LOGGER.debug("Reading the cached database from %s", str(cache_file))

        cls._read = True
        with open(cache_file, "rb") as cache:
            cast(Database, cls._instance).update(pickle.load(cache))

    @classmethod
    def save_cache(cls) -> None:
        """Saves the current database to a cache.

        This method does nothing, if the cache is already newer than the last modification time of
        the database file itself.
        """
        cache_file = cls._get_cache_file()
        if cache_file is None:
            return

        LOGGER.debug("Caching the database in %s", str(cache_file))

        try:
            cache_outdated = cls._is_cache_outdated(cache_file)
            if not cache_outdated:
                LOGGER.info("The cache is already up-to-date")
                return
        except FileNotFoundError:
            cache_file.parent.mkdir(parents=True, exist_ok=True)

        with open(cache_file, "wb") as cache:
            pickle.dump(cls._instance, cache)


class CacheError(Exception):
    """An error class used to handle cache events."""

    log_level: int = 35
    """The logging level at which to log this exception when it gets caught."""



"""coBib's UIs.

coBib is designed with two user interfaces in mind:
- a command-line interface (short: CLI)
- a terminal user interface (short: TUI)
"""

from .cli import CLI as CLI
from .tui import TUI as TUI
from .ui import UI as UI



"""coBib's Entry class."""

from __future__ import annotations

import logging
import subprocess
from itertools import accumulate
from typing import TYPE_CHECKING, Any, List, Optional, cast

from pylatexenc.latex2text import LatexNodes2Text
from pylatexenc.latexencode import UnicodeToLatexEncoder
from text_unidecode import unidecode

from cobib.config import AuthorFormat, config
from cobib.utils.match import Match, Span
from cobib.utils.regex import HAS_OPTIONAL_REGEX, regex
from cobib.utils.rel_path import RelPath

from .author import Author

if TYPE_CHECKING:
    import cobib.parsers

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class Entry:
    """coBib's bibliographic entry.

    coBib's `Database` stores the bibliographic information in entries which are instances of this
    class. This only contains a `label` which is a string used as a key to associate this entry as
    well as a free-form `data` dictionary, which can contain arbitrary key-value pairs.
    One field will *always* be present in the `data` dictionary:
    * `ENTRYTYPE`: which specifies the BibLaTex type of the entry.

    Only through the context imposed by BibLaTex will the other `data` fields be interpreted.

    However, some keys are exposed as properties in a special format to provide easy access to these
    meta-data fields not typically used by BibLaTex itself.

    Besides being a data container, this class also provides some utilities for data manipulation
    and querying.
    """

    _unicode_to_latex_encoder: UnicodeToLatexEncoder | None = None
    """The singleton `UnicodeToLatexEncoder` used by all instances of this class."""

    @classmethod
    def _get_unicode_to_latex_encoder(cls) -> UnicodeToLatexEncoder:
        """Returns the `_unicode_to_latex_encoder` singleton and constructs it if necessary."""
        if cls._unicode_to_latex_encoder is None:
            cls._unicode_to_latex_encoder = UnicodeToLatexEncoder(
                non_ascii_only=True,
                replacement_latex_protection="braces-all",
                unknown_char_policy="keep",
                unknown_char_warning=not config.database.format.suppress_latex_warnings
                or LOGGER.isEnabledFor(logging.DEBUG),
            )

        return cls._unicode_to_latex_encoder

    _latex_to_text_decoder: LatexNodes2Text | None = None
    """The singleton `LatexNodes2Text` used by all instances of this class."""

    @classmethod
    def _get_latex_to_text_decoder(cls) -> LatexNodes2Text:
        """Returns the `_latex_to_text_decoder` singleton and constructs it if necessary."""
        if cls._latex_to_text_decoder is None:
            cls._latex_to_text_decoder = LatexNodes2Text(keep_braced_groups=True)

        return cls._latex_to_text_decoder

    def __init__(self, label: str, data: dict[str, Any]) -> None:
        """Initializes a new Entry.

        Args:
            label: the label associated with this entry in the `Database`.
            data: the actual bibliographic data stored as a dictionary mapping free-form field names
                (`str`) to any other data. Some fields are exposed as properties of this class for
                convenience.
        """
        LOGGER.debug("Initializing entry: %s", label)

        self._label: str = str(label)

        self.data: dict[str, Any] = {}
        """The actual bibliographic data."""

        for key, value in data.items():
            if hasattr(self, key):
                setattr(self, key, value)
            elif isinstance(value, str) and value.isnumeric():
                LOGGER.info(
                    "Converting field '%s' of entry '%s' to integer: %s.",
                    key,
                    label,
                    value,
                    extra={"entry": label, "field": key},
                )
                self.data[key] = int(value)
            else:
                self.data[key] = value

        if "ID" in self.data:
            self.data.pop("ID")
            LOGGER.info(
                "The field '%s' of entry '%s' is no longer required. It will be inferred from the "
                "entry label.",
                "ID",
                label,
                extra={"entry": label, "field": "ID"},
            )

    def __eq__(self, other: object) -> bool:
        """Checks equality of two entries."""
        if not isinstance(other, Entry):
            return False
        return self.label == other.label and self.data == other.data

    def merge(self, other: Entry, *, ours: bool = False) -> None:
        """Merges the current entry with the other one.

        Args:
            other: the other entry.
            ours: a boolean indicating which data takes precedence. `True` indicates that `self`
                takes precedence over `other` which means that `other.update(self)` will be run
                (`self` overwrites values in `other`). `False` (the default) indicates the opposite.
        """
        if ours:
            other_data = other.data.copy()
            other_data.update(self.data)
            self.data = other_data.copy()
        else:
            self.data.update(other.data)

    @property
    def label(self) -> str:
        """The `Database` label of this entry."""
        return self._label

    @label.setter
    def label(self, label: str) -> None:
        """Sets the `Database` label of this entry.

        Args:
            label: the label of this entry.
        """
        LOGGER.debug("Changing the label '%s' to '%s'.", self.label, label)
        self._label = str(label)

    def markup_label(self) -> str:
        """Returns the label of this entry with the rich markup based on special tags."""
        markup_label = self.label

        markup_tags: dict[str, int] = {}
        for tag in self.tags:
            if tag in config.theme.tags.names:
                markup_tags[f"tag.{tag}"] = config.theme.tags.weights[tag]

        for tag, _ in sorted(markup_tags.items(), key=lambda item: item[1], reverse=True):
            markup_label = f"[{tag}]{markup_label}[/{tag}]"

        return markup_label

    @property
    def author(self) -> str:
        """The author of this entry formatted as a single string.

        Internally, the author(s) may be stored as one or more strings or `cobib.database.Author`
        instances. In this case, they will be joined into a single string of unified formatting
        before returned by this method.

        Returns:
            The string-formatted author(s) of this entry.

        References:
            - https://www.bibtex.com/f/author-field/
        """
        authors = self.data.get("author", "")

        if isinstance(authors, str):
            return authors

        if isinstance(authors, list):
            author_string = ""
            for author in authors:
                if author_string:
                    author_string += " and "
                author_string += str(author)
            return author_string

        return ""

    @author.setter
    def author(self, authors: str | list[str] | list[dict[str, str]] | list[Author]) -> None:
        """Sets the author of this entry.

        The input is processed in two steps.
        1. any encountered LaTeX commands are decoded into Unicode.
        2. each author is parsed by `cobib.database.Author.parse`.

        Args:
            authors: can be a single string, list of strings, list of `cobib.database.Author`
                objects, or a list of dictionaries mapping strings to strings which are treated as
                the key-value pairs to reconstruct a `cobib.database.Author` named tuple instance.

        References:
            - https://www.bibtex.com/f/author-field/
        """
        if not isinstance(authors, list):
            authors = authors.split(" and ")

        dec = self._get_latex_to_text_decoder()

        parsed_authors: list[str | Author] = []
        for author in authors:
            if isinstance(author, Author):
                parsed_authors.append(author)
                continue
            if isinstance(author, dict):
                parsed_authors.append(Author(**author))
                continue

            # we explicitly convert any latex instructions to Unicode
            author = cast(str, dec.latex_to_text(author))  # noqa: PLW2901
            LOGGER.debug("Converted the author to Unicode: '%s'", author)

            parsed_author = Author.parse(author)
            if config.database.format.author_format == AuthorFormat.YAML and not isinstance(
                parsed_author, str
            ):
                LOGGER.info(
                    "Parsed the author '%s' of entry '%s' from a string to the more detailed "
                    "information. You can consider storing it as such directly.",
                    author,
                    self.label,
                    extra={"entry": self.label, "field": "author"},
                )
            parsed_authors.append(parsed_author)

        self.data["author"] = parsed_authors

    @property
    def tags(self) -> list[str]:
        """The tags of this entry."""
        return cast(List[str], self.data.get("tags", []))

    @tags.setter
    def tags(self, tags: str | list[str]) -> None:
        """Sets the tags of this entry.

        Args:
            tags: a single or list of tags.
        """
        if isinstance(tags, list):
            self.data["tags"] = tags
        else:
            self.data["tags"] = tags.split(config.database.stringify.list_separator.tags)
            if len(self.data["tags"]) > 1:
                LOGGER.info(
                    "Converted the field '%s' of entry '%s' to a list. You can consider storing it "
                    "as such directly.",
                    "tags",
                    self.label,
                    extra={"entry": self.label, "field": "tags"},
                )
        LOGGER.debug("Adding the tags '%s' to '%s'.", self.data["tags"], self.label)

    @property
    def file(self) -> list[str]:
        """The associated files of this entry.

        The setter of this property will convert the strings to paths relative to the user's home
        directory. Internally, this field will always be stored as a list.
        """
        return cast(List[str], self.data.get("file", []))

    @file.setter
    def file(self, file: str | list[str]) -> None:
        """Sets the associated files of this entry.

        Args:
            file: can be either a single path (`str`) or a list thereof. In either case, the strings
                will be converted to paths relative to the user's home directory. Internally, this
                field will always be stored as a list.
        """
        if isinstance(file, list):
            paths = [RelPath(f) for f in file]
        else:
            paths = [RelPath(f) for f in file.split(config.database.stringify.list_separator.file)]
            if len(paths) > 1:
                LOGGER.info(
                    "Converted the field '%s' of entry '%s' to a list. You can consider storing it "
                    "as such directly.",
                    "file",
                    self.label,
                    extra={"entry": self.label, "field": "file"},
                )
        self.data["file"] = [str(p) for p in paths]
        LOGGER.debug("Adding '%s' as the file to '%s'.", self.data["file"], self.label)

    @property
    def url(self) -> list[str]:
        """The associated URL(s) of this entry."""
        return cast(List[str], self.data.get("url", []))

    @url.setter
    def url(self, url: str | list[str]) -> None:
        """Sets the associated URL(s) of this entry.

        Args:
            url: can be either a single URL (`str`) or a list thereof. Internally, this field will
            always be stored as a list.
        """
        if isinstance(url, list):
            self.data["url"] = url
        else:
            self.data["url"] = url.split(config.database.stringify.list_separator.url)
            if len(self.data["url"]) > 1:
                LOGGER.info(
                    "Converted the field '%s' of entry '%s' to a list. You can consider storing it "
                    "as such directly.",
                    "url",
                    self.label,
                    extra={"entry": self.label, "field": "url"},
                )
        LOGGER.debug("Adding '%s' as the url to '%s'.", self.data["url"], self.label)

    @property
    def month(self) -> str | None:
        """Returns the month."""
        return cast(Optional[str], self.data.get("month", None))

    @month.setter
    def month(self, month: int | str) -> None:
        """Sets the month."""
        months = [
            "jan",
            "feb",
            "mar",
            "apr",
            "may",
            "jun",
            "jul",
            "aug",
            "sep",
            "oct",
            "nov",
            "dec",
        ]
        if month in months:
            self.data["month"] = month
        else:
            if isinstance(month, int):
                self.data["month"] = months[month - 1]
            elif isinstance(month, str):
                if month.isnumeric():
                    self.data["month"] = months[int(month) - 1]
                else:
                    self.data["month"] = month.lower()[:3]
            LOGGER.info(
                "Converting field '%s' of entry '%s' from '%s' to '%s'.",
                "month",
                self.label,
                month,
                self.data["month"],
                extra={"entry": self.label, "field": "month"},
            )

    def stringify(self, *, encode_latex: bool = True, markup: bool = False) -> dict[str, str]:
        """Returns an identical entry to self but with all fields converted to strings.

        Args:
            encode_latex: whether to encode non-ASCII characters using LaTeX sequences. If this is
                `True`, a `pylatexenc.latexencode.UnicodeToLatexEncoder` will be used to replace
                Unicode characters with LaTeX commands.
            markup: whether or not to add markup based on the configured special tags.

        Returns:
            The data of this `Entry` as pure string fields.
        """
        if encode_latex:
            enc = self._get_unicode_to_latex_encoder()
        data = {}
        data["label"] = self.markup_label() if markup else self.label
        for field, value in self.data.items():
            if hasattr(self, field):
                value = getattr(self, field)  # noqa: PLW2901
            if field == "author":
                data[field] = str(value)
            elif isinstance(value, list) and hasattr(
                config.database.stringify.list_separator, field
            ):
                data[field] = getattr(config.database.stringify.list_separator, field).join(value)
            else:
                data[field] = str(value)
            if encode_latex:
                data[field] = enc.unicode_to_latex(data[field])
        return data

    def formatted(self) -> Entry:
        """Formats the entry in a clean and reproducible manner.

        This includes the following:

        1. escaping of special characters; these should be escaped to ensure proper rendering in
           LaTeX documents. This function leverages the existing implementation of the `pylatexenc`
           module to do said conversion. The only fields exempted from the conversion are those

        2. handles the `cobib.config.config.DatabaseFormatConfig.author_format` setting and outputs
           the `author` field in the corresponding format.

        Returns:
            A new `Entry` instance with all fields properly formatted.
        """
        enc = self._get_unicode_to_latex_encoder()
        formatted_entry = Entry(self.label, {})
        for key, value in self.data.items():
            if key in config.database.format.verbatim_fields:
                # do NOT these fields and keep any special characters
                formatted_entry.data[key] = value
                continue

            if key == "author":
                if config.database.format.author_format == AuthorFormat.BIBLATEX:
                    formatted_entry.data[key] = enc.unicode_to_latex(self.author)
                elif config.database.format.author_format == AuthorFormat.YAML:
                    formatted_entry.data[key] = value
                continue

            if isinstance(value, str):
                formatted_entry.data[key] = enc.unicode_to_latex(value)
            else:
                formatted_entry.data[key] = value

        return formatted_entry

    def save(self, parser: cobib.parsers.base_parser.Parser | None = None) -> str:
        """Saves an entry using the parsers `dump` method.

        This method is mainly used by the `Database.save` method and takes care of some final
        conversions depending on the user's configuration. Applying such modifications (like e.g.
        special character escaping) only before saving ensures a consistent state of the database
        while also providing a fast startup because these conversions are prevented at that time.

        Args:
            parser: the parser instance to use for dumping. If set to `None` it will default to a
                `cobib.parsers.YAMLParser`. Supplying a ready instance can improve efficiency
                significantly while saving many entries after one another.

        Returns:
            The string-representation of this entry as produced by the provided parser.
        """
        formatted_entry = self.formatted()
        if parser is None:
            from cobib.parsers.yaml import YAMLParser

            parser = YAMLParser()
        return parser.dump(formatted_entry) or ""  # `dump` may return `None`

    def matches(
        self,
        filter_: dict[tuple[str, bool], list[str]],
        or_: bool,
        *,
        ignore_case: bool = False,
        decode_unicode: bool = False,
        decode_latex: bool = False,
        fuzziness: int = 0,
    ) -> bool:
        """Check whether this entry matches the supplied filter.

        coBib provides an extensive filtering implementation. The filter is specified in the form
        of a dictionary whose keys consist of pairs of `(str, bool)` entries where the string
        indicates the field to match against and the boolean whether a positive (`true`) or negative
        (`false`) match is required. The values of the dictionary must be a `list[str]`. This means
        that field types are always compared on a string-basis (this is a limitation of the targeted
        command-line interface). However, as of v3.2.0 these strings are interpreted as regex
        patterns providing the most power within this framework.

        Some examples:

        | `filter_`                             | `or_`    | Meaning                               |
        | ------------------------------------- | -------- | ------------------------------------- |
        | `{('year', True): ['2020']}`          | *either* | `year` contains 2020                  |
        | `{('year', False): ['2020']}`         | *either* | `year` does not contain 2020          |
        | `{('year', True): ['2020', '2021']}`  | True     | `year` contains either 2020 or 2021   |
        | `{('year', True): ['2020', '2021']}`  | False    | cannot match anything                 |
        | `{('year', False): ['2020', '2021']}` | False    | `year` contains neither 2020 nor 2021 |

        Args:
            filter_: dictionary describing the filter as explained above.
            or_ : boolean indicating whether logical OR (`true`) or AND (`false`) is used to combine
                multiple filter items.
            ignore_case: if True, the matching will be case-*in*sensitive.
            decode_unicode: if True, all Unicode characters will be decoded before matching.
            decode_latex: if True, all LaTeX sequences will be decoded before matching.
            fuzziness: the amount of fuzzy errors to allow for matches. Using this feature requires
                the optional `regex` dependency to be installed.

        Returns:
            Boolean indicating whether this entry matches the filter.
        """
        LOGGER.debug("Checking whether entry %s matches.", self.label)

        if decode_latex:
            dec = self._get_latex_to_text_decoder()

        re_flags = regex.IGNORECASE if ignore_case else 0

        match_list = []
        stringified_data = self.stringify(encode_latex=False)

        for key, values in filter_.items():
            if key[0] not in stringified_data:
                match_list.append(not key[1])
                continue

            field_data = stringified_data[key[0]]

            if decode_latex:
                field_data = dec.latex_to_text(field_data)

            if decode_unicode:
                field_data = unidecode(field_data)

            for val in values:
                if fuzziness:
                    re_compiled = regex.compile(rf"({val}){{e<={fuzziness}}}", flags=re_flags)
                else:
                    re_compiled = regex.compile(rf"{val}", flags=re_flags)

                if re_compiled.search(field_data):
                    match_list.append(key[1])
                else:
                    match_list.append(not key[1])
        if or_:
            return any(m for m in match_list)
        return all(m for m in match_list)

    def search(  # noqa: PLR0912
        self,
        query: list[str],
        *,
        context: int = 1,
        skip_files: bool = False,
        ignore_case: bool = False,
        decode_unicode: bool = False,
        decode_latex: bool = False,
        fuzziness: int = 0,
    ) -> list[Match]:
        """Search entry contents for the query strings.

        The entry will *always* be converted to a searchable string using the
        `cobib.parsers.BibtexParser.dump` method.

        .. note::
           Setting `decode_latex=True` and/or `decode_unicode=True` does *NOT* affect the output of
           the above in order to ensure that the printed search results still include the original
           LaTeX sequences and/or Unicode characters. Instead, only the internally searched string
           will be decoded.

        This text will then be search for each item in `query` and will interpret these as regex
        patterns.
        If a `file` is associated with this entry, the search will try its best to recursively query
        its contents, too. However, the success of this depends highly on the configured search
        tool, `cobib.config.config.SearchCommandConfig.grep`.

        Args:
            query: the list of regex patterns to search for.
            context: the number of context lines to provide for each match. This behaves similarly
                to the *Context Line Control* available for the UNIX `grep` command (`--context`).
            skip_files: if True, associated files will *not* be searched.
            ignore_case: if True, the search will be case-*in*sensitive.
            decode_unicode: if True, all Unicode characters will be decoded before search.
            decode_latex: if True, all LaTeX sequences will be decoded before search.
            fuzziness: the amount of fuzzy errors to allow for search matches. Using this feature
                requires the optional `regex` dependency to be installed.

        Returns:
            A list of lists containing the context for each match associated with this entry.
        """
        if fuzziness > 0 and not HAS_OPTIONAL_REGEX:  # pragma: no branch
            LOGGER.warning(  # pragma: no cover
                "Using the `fuzziness` option requires the optional `regex` dependency to be "
                "installed! Falling back to `fuzziness=0`."
            )

        LOGGER.debug("Searching entry %s.", self.label)
        matches: list[Match] = []

        from cobib.parsers.bibtex import BibtexParser

        # get searchable text
        bibtex_raw = BibtexParser(encode_latex=False).dump(self)

        # split into lines and compute their lengths and offsets
        lines = bibtex_raw.split("\n")
        line_lengths = [len(line) + 1 for line in lines]  # + 1 to account for the newline character
        line_offsets = list(accumulate(line_lengths))

        if decode_latex:
            dec = self._get_latex_to_text_decoder()
            bibtex_raw = dec.latex_to_text(bibtex_raw)

        if decode_unicode:
            bibtex_raw = unidecode(bibtex_raw)

        re_flags = regex.IGNORECASE if ignore_case else 0
        for query_str in query:
            if fuzziness:
                re_compiled = regex.compile(rf"({query_str}){{e<={fuzziness}}}", flags=re_flags)
            else:
                re_compiled = regex.compile(rf"{query_str}", flags=re_flags)

            # find all query matches
            re_matches = list(re_compiled.finditer(bibtex_raw))

            # determine line index for each match
            matched_line_indices = []
            for match_ in re_matches:
                end = match_.span()[1]
                for line_idx, offset in enumerate(line_offsets):
                    if offset > end:
                        matched_line_indices.append(line_idx)
                        break

            matched_line_indices_set = set(matched_line_indices)
            prev_idx = 0
            for idx, (match_, line_idx) in enumerate(zip(re_matches, matched_line_indices)):
                try:
                    next_idx = matched_line_indices[idx + 1]
                except IndexError:
                    next_idx = len(lines)

                start = line_idx
                # NOTE: we iterate in reverse in order to ensure that we abort on the first previous
                # occurrence of the query pattern
                for idx2 in reversed(
                    range(
                        max(line_idx - context, prev_idx),
                        min(line_idx, next_idx),
                    )
                ):
                    if idx2 in matched_line_indices_set:
                        break
                    start -= 1

                stop = line_idx + 1
                for idx2 in range(
                    max(line_idx + 1, prev_idx),
                    min(line_idx + context + 1, next_idx),
                ):
                    if idx2 in matched_line_indices_set:
                        break
                    stop += 1

                offset = line_offsets[start - 1] if start > 0 else 0
                matches.append(
                    Match(
                        "\n".join(lines[start:stop]),
                        [Span(match_.start() - offset, match_.end() - offset)],
                    )
                )

                prev_idx = line_idx

            if skip_files:
                LOGGER.debug("Skipping the search in associated files of %s", self.label)
                continue

            for file_ in self.file:
                grep_prog = config.commands.search.grep
                path = RelPath(file_).path
                if not path.exists():
                    LOGGER.warning(
                        "The associated file %s of entry %s does not exist!", file_, self.label
                    )
                    continue

                LOGGER.debug("Searching associated file %s with %s", file_, grep_prog)
                with subprocess.Popen(
                    [
                        grep_prog,
                        *config.commands.search.grep_args,
                        f"-C{context}",
                        query_str,
                        RelPath(file_).path,
                    ],
                    stdout=subprocess.PIPE,
                ) as grep:
                    if grep.stdout is None:
                        continue
                    stdout = grep.stdout.read().decode()
                    if not stdout:
                        continue

                for file_match in stdout.split("\n--\n"):
                    stripped = file_match.strip()
                    file_matches = list(re_compiled.finditer(stripped))
                    matches.append(
                        Match(stripped, [Span(m.start(), m.end()) for m in file_matches])
                    )

        return matches



"""coBib's command-line interface.

This class provides the main entry point to coBib. It exposes all the commands implemented in
`cobib.commands` to the end-user and leverages [`rich`](https://github.com/textualize/rich) to
produce beautiful output.
"""

import argparse
import asyncio
import logging
import sys
from inspect import iscoroutinefunction
from typing import Any

from rich.console import Console
from typing_extensions import override

from cobib import __version__
from cobib.config import config
from cobib.ui.components.entry_points import entry_points
from cobib.ui.tui import TUI
from cobib.ui.ui import UI
from cobib.utils.logging import print_changelog

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class CLI(UI):
    """The CLI class.

    In addition to the global arguments documented by the base class, the following are supported:

      * `--version`: prints the coBib version and quits.
      * `command`: a single positional argument indicating the name of the command to run. When this
        is omitted, the `cobib.ui.tui.TUI` gets started.
    """

    @override
    def add_extra_parser_arguments(self) -> None:
        self.parser.add_argument("--version", action="version", version=f"%(prog)s v{__version__}")

        self.parser.add_argument(
            "command",
            help="the subcommand to be run",
            choices=sorted([cls.name for (cls, _) in entry_points("cobib.commands")]),
            nargs="?",
        )
        self.parser.add_argument(
            "args", nargs=argparse.REMAINDER, help="any arguments passed on to the subcommand"
        )

    @override
    def parse_args(self) -> argparse.Namespace:
        arguments = super().parse_args()
        if arguments.command is not None:
            sys_args = list(sys.argv)
            subcmd_args = sys_args[sys_args.index(arguments.command) + 1 :]
            if subcmd_args != arguments.args:
                LOGGER.log(
                    35,
                    "The arguments provided after the subcommand name did did not match the parsed "
                    "ones. This can occur in rare cases when the '--' pseudo-argument is involved. "
                    "Taking an educated guess and overwriting them. Please file a bug report if "
                    "this is a wrong assumption: https://gitlab.com/cobib/cobib/-/issues/new",
                )
                arguments.args = subcmd_args
        return arguments

    @override
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.init_argument_parser(
            prog="coBib",
            description=(
                "coBib input arguments.\nIf no subcommand is given, the TUI will be started."
            ),
        )

    @override
    async def run(self) -> None:  # type: ignore[override]
        arguments = self.parse_args()

        console = Console(theme=config.theme.build())

        if not arguments.porcelain:
            # print latest changelog
            changelog = print_changelog(__version__, config.logging.version)
            if changelog is not None:
                console.print(changelog)

        if not arguments.command:
            task = asyncio.create_task(TUI().run_async())  # type: ignore[abstract]
            await task
            # the following is required for the asynchronous TUI to quit properly
            sys.exit()
        else:
            (cls,) = [
                cls for (cls, _) in entry_points("cobib.commands") if cls.name == arguments.command
            ]
            subcmd = cls.load()(*arguments.args)
            if iscoroutinefunction(subcmd.execute):
                await subcmd.execute()
            else:
                subcmd.execute()
            if arguments.porcelain:
                output = subcmd.render_porcelain()
                for line in output:
                    print(line)
            else:
                renderable = subcmd.render_rich()
                if renderable is not None:
                    console.print(renderable)



"""coBib's help screen.

This screen renders a help information with the current key bindings and action descriptions.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from typing import ClassVar

from rich.table import Table
from textual.app import ComposeResult
from textual.binding import Binding
from textual.screen import ModalScreen
from textual.widgets import Static
from typing_extensions import override


class HelpScreen(ModalScreen[None]):
    """coBib's help screen."""

    BINDINGS: ClassVar[list[Binding | tuple[str, str] | tuple[str, str, str]]] = [
        ("question_mark", "toggle_help", "Help"),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | ? | Toggles the help screen. |
    """

    DEFAULT_CSS = """
        HelpScreen {
            align: center middle;
        }

        #help {
            padding: 1 2;
            width: auto;
            height: auto;
            background: $surface;
        }
    """

    HELP_DESCRIPTIONS: ClassVar[list[tuple[str, str]]] = [
        ("q", "Quit's coBib"),
        ("question_mark", "Toggles the help screen"),
        ("underscore", "Toggles between the horizontal and vertical layout"),
        ("space", "Toggles folding of a search result"),
        ("colon", "Starts the prompt for an arbitrary coBib command"),
        ("v", "Selects the current entry"),
        ("slash", "Searches the database for the provided string"),
        ("digit", "Immediately selects the preset filter given by that digit (0 = reset)"),
        ("a", "Prompts for a new entry to be added to the database"),
        ("d", "Deletes the current (or selected) entries"),
        ("e", "Edits the current entry"),
        ("f", "Allows filtering the table using `++/--` keywords"),
        ("i", "Imports entries from another source"),
        ("m", "Prompts for a modification (respects selection)"),
        ("o", "Opens the current (or selected) entries"),
        ("p", "Allows selecting a preset filter (see `config.tui.preset_filters`)"),
        ("r", "Redoes the last undone change. Requires git-tracking!"),
        ("s", "Prompts for the field to sort by (use -r to list in reverse)"),
        ("u", "Undes the last change. Requires git-tracking!"),
        ("x", "Exports the current (or selected) entries"),
        ("j", "Moves one row down"),
        ("k", "Moves one row up"),
        ("h", "Moves to the left"),
        ("l", "Moves to the right"),
    ]
    """The key binding help descriptions."""

    @override
    def compose(self) -> ComposeResult:
        help_table = Table(title="coBib TUI Help", caption="Close this help by pressing '?'")
        help_table.add_column("Key")
        help_table.add_column("Description")
        for key, description in self.HELP_DESCRIPTIONS:
            help_table.add_row(self.app.get_key_display(key), description)
        static = Static(help_table, id="help")
        static.styles.content_align = ("center", "middle")
        yield static

    def action_toggle_help(self) -> None:
        """Toggles the help information.

        Since this is the action of the `HelpScreen`, it simply pops the screen.
        """
        self.app.pop_screen()



"""coBib's Rich-Textual prompt integration.

The `Prompt` utility of this module either constructs a `rich.prompt.Prompt` or a custom textual
widget to prompt the user for input. In the latter case, it leverages coBib's
`cobib.ui.components.input_screen.InputScreen` for handling the input.

This implementation permits the seamless integration of interactive user prompts during command
execution with an interactive TUI based on textual.

The `Confirm` utility provides an even simpler interface for yes/no prompts.
"""

from __future__ import annotations

import asyncio
from functools import wraps
from typing import Any, Callable, TextIO, Type, cast

from rich.console import RenderableType
from rich.prompt import Confirm as RichConfirm
from rich.prompt import InvalidResponse, PromptBase, PromptType
from rich.prompt import Prompt as RichPrompt
from rich.text import Text, TextType
from textual.app import App
from textual.widgets import Input, Static
from typing_extensions import override

from .context import get_active_app


class Prompt:
    """A utility class to construct either a `rich` or `textual` prompt."""

    @staticmethod
    async def ask(
        prompt: TextType = "",
        *,
        choices: list[str] | None = None,
        default: Any = ...,
        show_choices: bool = True,
        show_default: bool = True,
        input_text: str = "",
        pre_prompt_message: RenderableType | None = None,
        process_response_wrapper: Callable[
            [Callable[[PromptBase[PromptType], str], PromptType]],
            Callable[[PromptBase[PromptType], str], PromptType],
        ]
        | None = None,
    ) -> Any:
        """Exposes the `rich.prompt.Prompt.ask` method.

        Args:
            prompt: the text for the prompt.
            choices: an optional list of choices.
            default: an optional default choice.
            show_choices: whether to display the choices (if any).
            show_default: whether to display the default choice.
            input_text: the text with which to pre-populate the input prompt.
            pre_prompt_message: a renderable message to display before the prompt.
            process_response_wrapper: a function to wrap the `rich.prompt.Prompt.process_response`
                method to allow the calling scope to inject response handling logic.
        """
        ask_prompt: Type[RichPrompt | TextualPrompt] = RichPrompt
        app = get_active_app()
        if app is not None:
            ask_prompt = TextualPrompt
            ask_prompt.input_text = input_text

        if process_response_wrapper is not None:
            ask_prompt.process_response = process_response_wrapper(  # type: ignore[assignment]
                ask_prompt.process_response  # type: ignore[arg-type]
            )

        if pre_prompt_message is not None:
            ask_prompt.pre_prompt = Prompt._wrap_pre_prompt(  # type: ignore[assignment]
                ask_prompt.pre_prompt, pre_prompt_message
            )

        if app is not None:
            res = await ask_prompt.ask(  # type: ignore[call-overload]
                prompt,
                choices=choices,
                default=default,
                show_choices=show_choices,
                show_default=show_default,
                console=app,
            )
        else:
            res = ask_prompt.ask(
                prompt,
                choices=choices,
                default=default,
                show_choices=show_choices,
                show_default=show_default,
            )

        if process_response_wrapper is not None:
            ask_prompt.process_response = (  # type: ignore[method-assign]
                ask_prompt.process_response.__wrapped__  # type: ignore[attr-defined]
            )

        if pre_prompt_message is not None:
            ask_prompt.pre_prompt = (  # type: ignore[method-assign]
                ask_prompt.pre_prompt.__wrapped__  # type: ignore[attr-defined]
            )

        return res

    @staticmethod
    def _wrap_pre_prompt(
        func: Callable[[PromptBase[PromptType]], None], message: RenderableType
    ) -> Callable[[PromptBase[PromptType]], None]:
        @wraps(func)
        def pre_prompt(prompt: PromptBase[PromptType]) -> None:
            app = get_active_app()
            if app is None:
                prompt.console.print(message)
            else:
                input_screen = app.get_screen("input")
                input_screen.mount(Static(message, id="panel"), before=-1)

        return pre_prompt


class Confirm:
    """A utility class to construct a yes/no `Prompt`."""

    @staticmethod
    async def ask(
        prompt: TextType = "",
        *,
        default: bool,
    ) -> bool:
        """Exposes the `rich.prompt.Confirm.ask` method.

        Args:
            prompt: the text for the prompt.
            default: the default answer.
        """
        confirm_prompt: Type[RichConfirm | TextualPrompt] = RichConfirm
        app = get_active_app()
        if app is not None:
            confirm_prompt = TextualPrompt

        confirm_prompt.process_response = Confirm._wrap_process_response(  # type: ignore[assignment]
            confirm_prompt.process_response  # type: ignore[arg-type]
        )

        if app is None:
            res = confirm_prompt.ask(prompt, default=default)
        else:
            res = await confirm_prompt.ask(  # type: ignore[call-overload]
                prompt, console=app, choices=["y", "n"], default="y" if default else "n"
            )

        confirm_prompt.process_response = (  # type: ignore[method-assign]
            confirm_prompt.process_response.__wrapped__  # type: ignore[union-attr]
        )

        return cast(bool, res)

    @staticmethod
    def _wrap_process_response(
        func: Callable[[PromptBase[bool], str], bool],
    ) -> Callable[[PromptBase[bool], str], bool]:
        @wraps(func)
        def process_response(prompt: PromptBase[bool], value: str) -> bool:
            return_value = func(prompt, value)

            if isinstance(return_value, bool):
                return return_value

            if cast(str, return_value).strip().lower() == "y":  # type: ignore[unreachable]
                return True

            if cast(str, return_value).strip().lower() == "n":
                return False

            raise InvalidResponse("You may only provide 'y' or 'n' as a valid response.")

        return process_response


class TextualPrompt(PromptBase[str]):
    """coBib's prompt widget."""

    input_text: str = ""
    """The text to pre-populate the input prompt with."""

    @override
    async def __call__(  # type: ignore[override]
        self, *, default: Any = ..., stream: TextIO | None = None
    ) -> str:
        while True:
            self.pre_prompt()
            prompt = self.make_prompt(default)
            value = await self.get_input(self.console, prompt, self.password, stream=stream)  # type: ignore[arg-type]
            if value == "" and default != ...:
                return str(default)
            try:
                return_value = self.process_response(value)
            except InvalidResponse as error:
                self.on_validate_error(value, error)
                continue
            else:
                return return_value

    @override
    @classmethod
    async def get_input(  # type: ignore[override]
        cls, console: App[None], prompt: TextType, password: bool, stream: TextIO | None = None
    ) -> str:
        dismiss_event = asyncio.Event()
        value = ""

        def _catch_dismissed_value(val: str) -> None:
            nonlocal value
            value = val
            dismiss_event.set()

        popup = Static(prompt + "\n", id="prompt")

        inp_screen = console.get_screen("input")
        # This is an cobib.ui.tui.components.InputScreen but casting to it would result in a
        # circular import. Instead, we ignore the mypy warning below.
        inp_screen.escape_enabled = False  # type: ignore[attr-defined]

        if inp_screen.is_current:
            # We need to pop the screen first to ensure that we can register our result callback.
            # One reason it might be the current screen already is if the pre_prompt has pushed it.
            console.pop_screen()

        await_mount = console.push_screen("input", _catch_dismissed_value)
        inp_screen.mount(popup, before=-1)
        await await_mount

        inp_screen.query_one(Input).value = cls.input_text
        await dismiss_event.wait()
        return value

    @override
    def on_validate_error(self, value: str, error: InvalidResponse) -> None:
        _id = "help" if value == "help" else "error"
        if isinstance(error.message, Text):
            message = Static(error.message, id=_id)
        else:
            message = Static(Text.from_markup(error.message), id=_id)

        self.console.get_screen("input").mount(message, before=-1)  # type: ignore[attr-defined]



"""coBib's Undo command.

This command can be used to undo the changes of the a previous command.
```
cobib undo
```
For obvious reasons, this will only undo commands which had an effect on the contents of the
database file.
Moreover, as a safety measure, this command will only undo those changes, which have been committed
by coBib *automatically*.
You can disable this be setting the `--force` argument which *always* undoes the last commit:
```
cobib undo --force
```

.. warning::
   This command is *only* available if coBib's git-integration has been enabled via
   `cobib.config.config.DatabaseConfig.git` *and* initialized properly (see `cobib.commands.init`).

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `u` key.
"""

from __future__ import annotations

import argparse
import logging
import subprocess
import sys
from pathlib import Path

from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.git import is_inside_work_tree
from cobib.utils.rel_path import RelPath

from .base_command import Command

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class UndoCommand(Command):
    """The Undo Command.

    This command can parse the following arguments:

        * `-f`, `--force`: if specified, this will also revert changes which have *not* been
          auto-committed by coBib.
    """

    name = "undo"

    @override
    def __init__(self, *args: str) -> None:
        super().__init__(*args)

        self.root: Path
        """The path to the root of the git repository tracking the database."""

        self.sha: str
        """The git commit SHA which was reverted by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = argparse.ArgumentParser(
            prog="undo", description="Undo subcommand parser.", exit_on_error=True
        )
        parser.add_argument(
            "-f", "--force", action="store_true", help="allow undoing non auto-committed changes"
        )
        cls.argparser = parser

    @override
    def execute(self) -> None:
        git_tracked = config.database.git
        if not git_tracked:
            msg = (
                "You must enable coBib's git-tracking in order to use the `Undo` command."
                "\nPlease refer to the documentation for more information on how to do so."
            )
            LOGGER.error(msg)
            return

        file = RelPath(config.database.file).path
        self.root = file.parent
        if not is_inside_work_tree(self.root):
            msg = (
                "You have configured, but not initialized coBib's git-tracking."
                "\nPlease consult `cobib init --help` for more information on how to do so."
            )
            LOGGER.error(msg)
            return

        LOGGER.debug("Starting Undo command.")

        Event.PreUndoCommand.fire(self)

        LOGGER.debug("Obtaining git log.")
        lines = subprocess.check_output(
            [
                "git",
                "--no-pager",
                "-C",
                f"{self.root}",
                "log",
                "--oneline",
                "--no-decorate",
                "--no-abbrev",
            ]
        )
        undone_shas = set()
        for commit in lines.decode().strip().split("\n"):
            LOGGER.debug("Processing commit %s", commit)
            self.sha, *message = commit.split()
            if message[0] == "Undo":
                # Store already undone commit sha
                LOGGER.debug("Storing undone commit sha: %s", message[-1])
                undone_shas.add(message[-1])
                continue
            if self.sha in undone_shas:
                LOGGER.info("Skipping %s as it was already undone", self.sha)
                continue
            if self.largs.force or (message[0] == "Auto-commit:" and message[-1] != "InitCommand"):
                # we undo a commit if and only if:
                #  - the `force` argument is specified OR
                #  - the commit is an `auto-committed` change which is NOT from `InitCommand`
                LOGGER.debug("Attempting to undo %s.", self.sha)
                commands = [
                    f"git -C {self.root} revert --no-commit {self.sha}",
                    f"git -C {self.root} commit --no-gpg-sign --quiet --message 'Undo {self.sha}'",
                ]
                with subprocess.Popen(
                    "; ".join(commands), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                ) as undo:
                    undo.communicate()
                    if undo.returncode != 0:
                        LOGGER.error(  # pragma: no cover
                            "Undo was unsuccessful. Please consult the logs and git history of your"
                            " database for more information."
                        )
                    else:
                        # update Database
                        Database().read()
                break
        else:
            msg = "Could not find a commit to undo. Please commit something first!"
            LOGGER.warning(msg)
            sys.exit(1)

        Event.PostUndoCommand.fire(self)



"""coBib's `logging.Handler` for interactive UIs.

This handler is used by the `cobib.ui.tui.TUI` to render logging messages inside of the
`cobib.ui.components.log_screen.LogScreen`.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast

from rich.text import Text
from typing_extensions import override

from .log_screen import LogScreen

if TYPE_CHECKING:
    from ..tui import TUI


class LoggingHandler(logging.Handler):
    """coBib's `logging.Handler` for interactive UIs."""

    def __init__(self, app: TUI, level: int = logging.INFO) -> None:
        """Initializes the handler.

        Args:
            app: the running TUI instance.
            level: the default logging level to be displayed.
        """
        super().__init__(level=level)

        self._app = app

        formatter = logging.Formatter(
            fmt="%(asctime)s [%(levelname)s] %(message)s", datefmt="%H:%M:%S"
        )
        self.setFormatter(formatter)

        for handler in self._app.root_logger.handlers[:]:
            if isinstance(handler, logging.StreamHandler):
                self._app.root_logger.removeHandler(handler)
                handler.close()

        self._app.root_logger.addHandler(self)

    @override
    def format(self, record: logging.LogRecord) -> Text:  # type: ignore[override]
        message = super().format(record)

        style = ""
        if record.levelno >= logging.CRITICAL:
            style = "bold red"
        elif record.levelno >= logging.ERROR:
            style = "red"
        elif record.levelno >= logging.WARNING:
            style = "yellow"
        elif record.levelno >= logging.INFO:
            style = "green"
        elif record.levelno >= logging.DEBUG:
            style = "blue"

        text = Text(message, style)
        return text

    @override
    def emit(self, record: logging.LogRecord) -> None:
        log_screen = cast(LogScreen, self._app.get_screen("log"))
        log_screen.rich_log.write(self.format(record))
        if record.levelno >= logging.ERROR and not log_screen.is_current:
            self._app.push_screen("log")
